# FILE: defaults.tcl
#
# This is a tcl script which performs the configuration task for the
# script ../build.tcl .  It sets values of Tcl variables which describe 
# your computing environment.  
#
# This script is not yet very sophisticated.  It is not able to learn 
# everything it needs to know about your computing environment on its 
# own.  It calls on two other scripts ./local.tcl and ./config.tcl, 
# which must contain definitions for variables which this script cannot
# determine on its own.  You will need to provide some of these 
# definitions by editing the files ./local.tcl and ./config.tcl according
# to the instructions in ./config.tcl .  You should not need to edit
# this file (./defaults.tcl).
#
######################################################################
# First, consult any saved definitions which describe the local platform.
# Consult them first so that they take priority.
set localscript [file dirname [info script]]/local.tcl
if [file exists $localscript] {
  EvaluateScript $localscript
}

######################################################################
# Next, consult any definitions in the (edited) configuration file.
EvaluateScript [file dirname [info script]]/config.tcl

######################################################################
# Finally, try to provide default values for everything, to cover any
# variables not defined in one of the above scripts.

# ------------- Platform description 
# Find out what machine we're building on
set platform_vars {PLATFORM MACHINE OS OSVERSION}

if [AllDefined tcl_platform] {
    Default PLATFORM [string tolower [string trim $tcl_platform(platform)]]
    Default MACHINE [string tolower [string trim $tcl_platform(machine)]]
    Default OS [string tolower [string trim $tcl_platform(os)]]
    Default OSVERSION [string tolower [string trim $tcl_platform(osVersion)]]
}
if {![AllDefined $platform_vars]} {
    ReportMissingValues $platform_vars
}
set OSMAJOR $OSVERSION
regexp {([^.]*).*} $OSVERSION match OSMAJOR

# ------------- System-level commands
set system_vars {MKDIR RM MV CP}

if { $tcl_version > 7.5 } {
    Default MKDIR	"file mkdir"
    Default RM		"file delete"
    Default MV		"file rename"
    Default CP		"file copy"
} else {
  if { [string compare $PLATFORM unix]==0 } {
    Default MKDIR	"ShellPrompt mkdir"
    Default RM		"ShellPrompt rm"
    Default MV		"ShellPrompt mv"
    Default CP		"ShellPrompt cp"
  } elseif {[string match windows $PLATFORM]} {
    Default MKDIR	"ShellPrompt mkdir"
    Default RM		"DOSRM"
    Default MV		"ShellPrompt move"
    Default CP		"DOSCP"
    } else {
# For other platforms, we need to fill in some configuration defaults.
  }
}
if {![AllDefined $system_vars]} {
    ReportMissingValues $system_vars
}

# ------------- Compiler/Linker configuration
set complink_vars {	CCC CCCEXT OBJEXT EXEEXT COMPILEONLYFLAG 
			MACRODEFINEFLAG INCLUDEDIRECTORYFLAG DEBUGFLAG 
			STRICTWARNFLAG OUTFLAG LINKOUTFLAG
			LIBRARYDIRECTORYFLAG LIBRARYFLAG
                        BROKENCCCERRORS LINKER
                  }
# Note: If you have a compiler that doesn't return a proper error code,
#   then you can set BROKENCCCERRORS to 1 and the compilation update
#   code will ignore the compiler error return.  An error will still be
#   raised if no object file is created as a result of the compilation,
#   however.

# Select platform/system specific defaults
if { [string compare $OS irix64]==0 } {
    # 64-bit SGI 
    Default CCC     "CC -32"
} elseif { [string compare $OS sunos]==0 && $OSMAJOR<5 } {
    # SunOS
    Default CCC "CC"
} elseif { [string compare $OS sunos]==0 && $OSMAJOR>=5 } {
    # Solaris
    Default CCC "CC -xwe"
    Default RANDOM_GENERATOR_HEADER_BLOCK {
/* Random number generator.  Replace with your own if you don't like    */
/* random().  Prototypes for random() and srandom() appear to be        */
/* missing from the Solaris system header files (though they are in the */
/* runtime library).  If you don't have random(), you can use rand()    */
/* instead, but be wary of some older, badly broken versions of rand(). */
extern "C" {
   /* If you get error messages about multiple inconsistent definitions  */
   /* of random or srandom, just yank out this prototype block.  (Better */
   /* yet, fix it up in your OOMMF/build/local.tcl file, as oommfport.h  */
   /* get rewritten whenever you do a 'build.tcl configure'              */
   long random(void);
   void srandom(unsigned int seed);
}
#define OMF_SRANDOM(seed)  srandom(seed)
#define OMF_RANDOM()       random()
#define OMF_RANDOM_MAX     0x7FFFFFFF /* Is this system dependent? */
    }
} elseif { [string compare $OS linux]==0 } {
    # Linux
    Default CCC "gcc"
    Default DLLIBRARYFILE libdl.so
} elseif { [string compare $PLATFORM windows]==0 \
        && [string compare $MACHINE intel]==0} {
    # Windows/intel
    Default CCC "cl"
Default RANDOM_GENERATOR_HEADER_BLOCK {
/* Random number generator. Replace with your own if */
/* if you don't like rand().                        */
#define OMF_SRANDOM(seed)  srand(seed)
#define OMF_RANDOM()       rand()
#define OMF_RANDOM_MAX     RAND_MAX
}
}
# Defaults for anything not set in one of the above blocks
Default CCC "gcc"
Default RANDOM_GENERATOR_HEADER_BLOCK {
/* Random number generator.  Replace with your own if */
/* you don't like random().                           */
#define OMF_SRANDOM(seed)  srandom(seed)
#define OMF_RANDOM()       random()
#define OMF_RANDOM_MAX     0x7FFFFFFF /* Is this system dependent? */
}

# Collect any recommended CFLAGS
Default CFLAGS {}

switch [lindex $CCC 0] {
  "CC" {Default CCCEXT 			".cc"
	Default OBJEXT 			".o"
	Default EXEEXT			""
	Default COMPILEONLYFLAG		"-c"
	Default MACRODEFINEFLAG		"-D"
	Default INCLUDEDIRECTORYFLAG	"-I"
	Default DEBUGFLAG		"-g"
	Default STRICTWARNFLAG		"+w"
	Default OUTFLAG 		"-o "	;# Trailing space needed
	Default LINKOUTFLAG 		"-o "	;# Trailing space needed
	Default LIBRARYDIRECTORYFLAG	"-L"
	Default LIBRARYFLAG		"-l"
	Default BROKENCCCERRORS		0
      Default LINKER			[lindex $CCC 0]
       }
  "gcc" {Default CCCEXT 			".cc"
	Default OBJEXT 			".o"
	Default EXEEXT			""
	Default COMPILEONLYFLAG		"-c"
	Default MACRODEFINEFLAG		"-D"
	Default INCLUDEDIRECTORYFLAG	"-I"
	Default DEBUGFLAG		"-g"
#        Default STRICTWARNFLAG		"-Wall"
        Default STRICTWARNFLAG "-Wall -W -Wpointer-arith \
		-Wbad-function-cast -Wwrite-strings \
		-Wstrict-prototypes -Wmissing-prototypes \
		-Wmissing-declarations -Wnested-externs -Winline \
		-Woverloaded-virtual -Wsynth -Werror"
# I would like to also include -Wshadow, -Wredundant-decls and
# -Wcast-align, but unclean system header files prevent this.  I would
# also like to use -Wcast-qual, but casting away const is needed on
# some occasions to provide "conceptual const" functions in place of
# "bitwise const"; cf. p76-78 of Meyer's book, "Effective C++."
	AppendUnique CFLAGS		"-O"
# Note: There is a unfortunate bug in some GCC's:
#  There is a library function 'const char *strchr(const char *,int)',
#  and an inlined function 'char *strchr(char *,int)', that calls
#  the former with casts.  Now, if optimization is turned off, these
#  problematic versions of gcc turn off inlining, and as a result
#  calls to 'char *strchr(char *,int)' will fail at the link step
#  because there is no such library routine.  Sigh...  This problem
#  also affects several other string functions.  See the header file
#  /usr/lib/g++-include/std/cstring.h.
	Default OUTFLAG 		"-o "	;# Trailing space needed
	Default LINKOUTFLAG 		"-o "	;# Trailing space needed
	Default LIBRARYDIRECTORYFLAG	"-L"
	Default LIBRARYFLAG		"-l"
	Default BROKENCCCERRORS		0
      Default LINKER			[lindex $CCC 0]
       }
  "cl" {Default CCCEXT 			".cpp"
	Default OBJEXT 			".obj"
	Default EXEEXT			".exe"
	Default COMPILEONLYFLAG		"/c"
	Default MACRODEFINEFLAG		"/D"
	Default INCLUDEDIRECTORYFLAG	"/I"
	Default DEBUGFLAG		""
	Default STRICTWARNFLAG		""
	Default OUTFLAG 		"/Fo"		;# *NO* trailing space
	Default LINKOUTFLAG 		"/OUT:"	;# *NO* trailing space
	Default LIBRARYDIRECTORYFLAG	""
	Default LIBRARYFLAG		""
	Default BROKENCCCERRORS		0
      Default LINKER			"link"
       }
}

if {![AllDefined $complink_vars]} {
  ReportMissingValues $complink_vars
}

# ------------- Compiler/Linker options
set option_vars {CFLAGS LFLAGS LIBXTR}

AppendUnique CFLAGS $COMPILEONLYFLAG

# Request strict compiler warnings
AppendUnique CFLAGS $STRICTWARNFLAG

# Construct CFLAGS from configuration information
# Inform compiler how to find Tcl/Tk header files
set includedirs {}
foreach dir {TCLINCLUDEDIRECTORY TKINCLUDEDIRECTORY X11INCLUDEDIRECTORY} {
  AppendUniqueExistingReadableDirectory includedirs $dir
}
foreach dir $includedirs {
  AppendUnique CFLAGS [list $INCLUDEDIRECTORYFLAG$dir]
}

# Construct LFLAGS from configuration information
Default LFLAGS {}

if {[string compare $CCC cl]==0} {
#  AppendUnique LFLAGS /link
#  AppendUnique LFLAGS "/NODEFAULTLIB /RELEASE /NOLOGO"
#  AppendUnique LFLAGS "/ALIGN:0x1000 /MACHINE:IX86"
#  AppendUnique LFLAGS "/SUBSYSTEM:WINDOWS /ENTRY:WinMainCRTStartup"
  # Need some defaults here
  Default TCLLIBRARYDIRECTORY "/tcl/bin"
  Default TKLIBRARYDIRECTORY "/tcl/bin"
  regsub {[.]} $tcl_version {} tcl_number
  Default TCLLIBRARYFILE tcl$tcl_number.lib
  # Works for Tcl 7.4-7.6 and 8.0+. Will still fail on 7.10
  if {$tcl_number > 79} {
    set tk_number $tcl_number
  } else {
    set tk_number [expr $tcl_number - 34]
  }
  Default TKLIBRARYFILE tk$tk_number.lib
# set tklibrary \"$TKLIBRARYDIRECTORY/$TKLIBRARYFILE\"
  set tklibrary $TKLIBRARYDIRECTORY/$TKLIBRARYFILE
  regsub -all {/} $tklibrary {\\} tklibrary
  AppendUnique LFLAGS [list $tklibrary]
# set tcllibrary \"$TCLLIBRARYDIRECTORY/$TCLLIBRARYFILE\"
  set tcllibrary $TCLLIBRARYDIRECTORY/$TCLLIBRARYFILE
  regsub -all {/} $tcllibrary {\\} tcllibrary
  AppendUnique LFLAGS [list $tcllibrary]
#  AppendUnique LFLAGS foo.lib
#  AppendUnique LFLAGS msvcrt.lib
#  AppendUnique LFLAGS libc.lib
#  AppendUnique LFLAGS oldnames.lib
#  AppendUnique LFLAGS kernel32.lib
#  AppendUnique LFLAGS advapi32.lib
  AppendUnique LFLAGS user32.lib
#  AppendUnique LFLAGS gdi32.lib
#  AppendUnique LFLAGS comdlg32.lib
#  AppendUnique LFLAGS winspool.lib
} else {
  # No known counterexamples for these two library names
  # Set defaults here (extension not really used)
  Default X11LIBRARYFILE "libX11.a"
  Default MATHLIBRARYFILE "libm.a"
  # A good guess for Tcl/Tk libraries
  Default TCLLIBRARYFILE "libtcl.a"
  Default TKLIBRARYFILE "libtk.a"
  # Try some guesses for the X11 lib path
  if {![info exists X11LIBRARYDIRECTORY]} {
    foreach try { "/usr/X11/lib" "/usr/X11R6/lib" "/usr/openwin/lib" } {
      if {[file isdirectory $try] && [file readable $try]} {
        set X11LIBRARYDIRECTORY $try
	break
      }
    }
  }

  set libdirs {}
  foreach dir {TCLLIBRARYDIRECTORY TKLIBRARYDIRECTORY X11LIBRARYDIRECTORY} {
    AppendUniqueExistingReadableDirectory libdirs $dir
  }
  foreach dir $libdirs {
    AppendUnique LFLAGS [list $LIBRARYDIRECTORYFLAG$dir]
  }
#
# This order is important!!!!!
#
  foreach lib {TKLIBRARYFILE TCLLIBRARYFILE DLLIBRARYFILE X11LIBRARYFILE 
          MATHLIBRARYFILE } {
    if [info exists $lib] {
      if [regexp {^lib(.*)[.][^.]*$} [set $lib] match libab] {
        AppendUnique LFLAGS [list $LIBRARYFLAG$libab]
      }
    }
  }
}

# By default, give the -ldl library flag a try

if {[string compare $CCC cl]} {
  Default LIBXTR "-ldl"
} else {
  Default LIBXTR ""
}

if {![AllDefined $option_vars]} {
  ReportMissingValues $option_vars
}

# ------------- Installation configuration
set install_vars {INSTALLROOT INSTALLEXECROOT INSTALLBIN INSTALLTCLLIB}

# Default installation directories
Default INSTALLROOT "/usr/local/oommf"
Default INSTALLEXECROOT $INSTALLROOT
Default INSTALLBIN "$INSTALLEXECROOT/bin"
Default INSTALLTCLLIB "$INSTALLROOT/lib"
Default INSTALLDEMOS "$INSTALLROOT/demos"

if {![AllDefined $install_vars]} {
  ReportMissingValues $install_vars
}


