# FILE: dialog.tk
#
# Tcl/Tk scripts for various dialog boxes
#
# Requires Tcl/Tk script file: misc.tk
#  It must be source'd _before_ this script.
#

###################### OMF_PrintDialog ###################################
# 4-Feb-1997, mjd
#
# "Local" global variables are stored in array OMFpd$id(), where id is
# obtained by converting .'s to _'s in the $window value passed
# to OMF_PrintDialog

# Utility proc to derive a printing output filename from an
# input name.  Currently just strips any extension, and
# appends ".ps".
proc OMF_PrintDialogDefaultFilename { inname } {
    return "[file rootname [string trim $inname]].ps"
}

proc OMF_PrintDialogKill { window } {
    OMF_SetFocus . ;# Transfer focus back to root window
                   ## before destroying current window
    OMF_GetID $window id base
    destroy $base   ;# Important: You will be unable to delete
                     # $printvar until _after_ $base (or at least
                     # the entry widget tied to $printvar) is
                     # destroyed!
    set printvar "OMFpd$id"
    global $printvar
    trace vdelete $printvar w OMF_PrintDialogTrace
    global UpdateViewCount
    trace vdelete UpdateViewCount w OMF_PrintDialogViewportTrace$id
    unset $printvar
}

# Determines width:height ratio using values of 
# pvar(canvas) and pvar(crop)
proc OMF_PrintDialogSetSizeRatio { pvarname } {
    upvar #0 $pvarname pvar
    if {$pvar(crop)} {
        # Printing cropped (viewable) region
        set width [$pvar(canvas) cget -width]
        set height [$pvar(canvas) cget -height]
    } else {
        # Printing whole canvas
        set sr [.canvas_frame.canvas cget -scrollregion]
        set width [expr [lindex $sr 2] - [lindex $sr 0]]
        set height [expr [lindex $sr 3] - [lindex $sr 1]]
    }
    if {$width<=0}  { set width 1. }   ;# Safety: Both ratio &
    if {$height<=0} { set height 1. }  ;# 1/ratio should be valid.
    global $pvarname
    set ${pvarname}(sizeratio) [expr double($width)/double($height)]
    # We need to access pvar(sizeratio) as ${pvarname}(sizeratio)
    # in order to get the trace on ${pvarname} to work properly.
}

# Initialize OMF_PrintDialog data array from outside print array
proc OMF_PrintDialogCopyIn { window parrname } {
    OMF_GetID $window id
    set pvarname "OMFpd$id"
    upvar #0 $pvarname pvar
    upvar #0 $parrname parr
    set pvar(window)    $window
    set pvar(canvas)    $parr(canvas)
    set pvar(inname)    $parr(inname)
    set pvar(outname)   $parr(outname)
    set pvar(orient)    $parr(orient)
    set pvar(papertype) $parr(papertype)
    set pvar(crop)      $parr(crop)
    # Position is broken into two variables inside OMF_PrintDialog
    set pvar(verpos) {} ;# Default is centered
    regexp -nocase {[ns]} $parr(position) pvar(verpos)
    set pvar(horpos) {} ;# Default is centered
    regexp -nocase {[ew]} $parr(position) pvar(horpos)
    # Size is handled differently internal to OMF_PrintDialog:
    set pvar(units) "i"  ;# Safety
    regexp -nocase {[ic]} $parr(xmargin) pvar(units)
    set pvar(oldunits) $pvar(units) ;# Used to track unit conversions
    set pvar(leftmargin) [OMF_ConvertUnits $parr(xmargin) $pvar(units)]
    set pvar(topmargin) [OMF_ConvertUnits $parr(ymargin) $pvar(units)]
    OMF_PrintDialogSetSizeRatio $pvarname
    set direction width; set length 1i  ;# Safety
    if ![regexp -- {-page(width|height) +([0-9.]+.)} $parr(size) match \
            direction length] {
        OMF_FatalError "Invalid print size switch ($parr(size)) detected" \
                "OMF_PrintDialogCopyIn Error"
    }
    if {"$direction"=="width"} {
        set width  [OMF_ConvertUnits $length $pvar(units)]
        set height [expr $width/$pvar(sizeratio)]
    } else {
        set height [OMF_ConvertUnits $length $pvar(units)]
        set width  [expr $height*$pvar(sizeratio)]
    }

    # Strip extra precision off of width & height
    set width [string trim $width]
    if {[string compare $width "."]==0} {
        set width "0."
    }
    if [regexp {\....} $width] {
        set width [format "%.2f" $width]
    }

    set height [string trim $height]
    if {[string compare $height "."]==0} {
        set height "0."
    }
    if [regexp {\....} $height] {
        set height [format "%.2f" $height]
    }

    set pvar(width)  $width
    set pvar(height) $height

    # If outname doesn't specify a pipe, build a default name
    # from inname
    if { "|" != [string index [string trim $pvar(outname)] 0] } {
	# Not a pipe
	set pvar(outname) [OMF_PrintDialogDefaultFilename $pvar(inname)]
    }

}

# Copy OMF_PrintDialog data array to outside print array
proc OMF_PrintDialogCopyOut { window parrname } {
    if { ![winfo exists $window] } { return }  ;# Window already destroyed
    OMF_GetID $window id
    set printvar "OMFpd$id"
    upvar #0 $printvar pvar
    upvar #0 $parrname parr
    set parr(inname)    $pvar(inname)
    set parr(outname)   $pvar(outname)
    set parr(orient)    $pvar(orient)
    set parr(papertype) $pvar(papertype)
    set parr(crop)      $pvar(crop)
    # Position is broken into two variables inside OMF_PrintDialog
    set parr(position)  "$pvar(verpos)$pvar(horpos)"
    if {"$parr(position)"==""} { set parr(position) "c" }

    # Size is handled differently internal to OMF_PrintDialog:
    set width [string trim $pvar(width)]
    if {[string compare $width ""]==0 \
        || [string compare $width "."]==0} {
        set width 0.
    }
    set parr(size)  "-pagewidth $width$pvar(units)"

    set leftmargin [string trim $pvar(leftmargin)]
    if {[string compare $leftmargin ""]==0 \
        || [string compare $leftmargin "."]==0} {
        set leftmargin 0.
    }
    set parr(xmargin) "$leftmargin$pvar(units)"

    set topmargin [string trim $pvar(topmargin)]
    if {[string compare $topmargin ""]==0 \
        || [string compare $topmargin "."]==0} {
        set topmargin "0."
    }
    set parr(ymargin) "$topmargin$pvar(units)"
}

# Fix up variables that depend on viewport display
proc OMF_PrintDialogUpdateVariables { pvarname } {
    OMF_PrintDialogSetSizeRatio $pvarname
}

# Trace procedure to automatically update PrintDialog variables
proc OMF_PrintDialogTrace { name elt op } {
    global $name
    if ![info exists ${name}($elt)] return
    # Disable tracing because we will be modifying several
    # elements of pvar.  Note: We must use the name "pvar"
    # inside the 'trace vdelete' command instead of "$name"
    # because that is how it is accessed inside this routine.
    upvar #0 $name pvar  ;# The #0 is needed for values
                          # modified inside entry widgets.
    trace vdelete pvar w OMF_PrintDialogTrace
    switch -exact -- $elt {
        sizeratio -
        width   {
            set temp [OMF_MakeValidPosNumber $pvar(width)]
            set pvar(width) $temp
            if {[string compare $temp ""]==0 \
                    || [string compare $temp "."]==0 } {
                set pvar(height) ""
            } else {
                set pvar(height) [expr $temp/$pvar(sizeratio)]
            }
        }
        height  {
            set temp [OMF_MakeValidPosNumber $pvar(height)]
            set pvar(height) $temp
            if {[string compare $temp ""]==0 \
                    || [string compare $temp "."]==0 } {
                set pvar(width) ""
            } else {
                set pvar(width) [expr $temp*$pvar(sizeratio)]
            }
        }
        crop    {
            OMF_PrintDialogSetSizeRatio $name
            set temp [OMF_MakeValidPosNumber $pvar(width)]
            set pvar(width) $temp
            if {[string compare $temp ""]==0 \
                    || [string compare $temp "."]==0 } {
                set pvar(height) ""
            } else {
                set pvar(height) [expr $temp/$pvar(sizeratio)]
            }
        }
        units {
            set pvar(units) [string trim $pvar(units)] ;# Safety
            if {[string compare $pvar(units) $pvar(oldunits)]!=0} {
                set pvar(leftmargin) [OMF_ConvertUnits \
                        "$pvar(leftmargin)$pvar(oldunits)" $pvar(units)]
                set pvar(topmargin) [OMF_ConvertUnits \
                        "$pvar(topmargin)$pvar(oldunits)" $pvar(units)]
                set pvar(width) [OMF_ConvertUnits \
                        "$pvar(width)$pvar(oldunits)" $pvar(units)]
                set pvar(height) [OMF_ConvertUnits \
                        "$pvar(height)$pvar(oldunits)" $pvar(units)]
                set pvar(oldunits) $pvar(units)
            }
        }
        default {}
    }
    # Strip extra precision off of width & height
    if [regexp {\....} $pvar(width)] {
        set pvar(width) [format "%.2f" $pvar(width)]
    }
    if [regexp {\....} $pvar(height)] {
        set pvar(height) [format "%.2f" $pvar(height)]
    }

    # Re-enable trace
    trace variable pvar w OMF_PrintDialogTrace
}


proc OMF_PrintDialogOK { window parrname } {
    if { ![winfo exists $window] } { return }  ;# Window already destroyed
    OMF_GetID $window id
    set pvarname "OMFpd$id"
    upvar #0 $pvarname pvar
    upvar #0 $parrname parr

    # Don't allow exit if outname is {}.  Instead
    # reset outname to default name derived from inname,
    # and return.
    if { [string match {} $pvar(outname)] } {
	set pvar(outname) [OMF_PrintDialogDefaultFilename $pvar(inname)]
	$window.file.value xview end   ;# Make end of outname viewable
	return
    }

    # Otherwise, copy pvar to parr and kill Print Dialog
    OMF_PrintDialogCopyOut $window $parrname
    set parr(done) "OK"
    OMF_PrintDialogKill $window
    # Note: The outside trace on parr(done) is removed when the
    # PrintDialog window is destroyed, so it is necessary to write
    # "OK" to parr(done) *before* calling OMF_PrintDialogKill.
}

proc OMF_PrintDialogCancel { window parrname } {
    if { ![winfo exists $window] } { return }  ;# Window already destroyed
    upvar #0 $parrname parr
    set parr(done) "Cancel"
    OMF_PrintDialogKill $window
}

# Here $parrname must be a global variable
proc OMF_PrintDialog { window title parrname } {
    toplevel $window
    wm group $window .
    wm title $window "$title"
    OMF_PositionChild $window  ;# Position at +.25+.25 over '.'
    upvar #0 $parrname parr
    # parr is a *global* "Printer Array" structure:
    #    canvas     -   name of canvas widget
    #    inname     -   name of file being displayed
    #    outname    -   name of output file or pipe
    #    orient     -   portrait | landscape
    #    size       -   output width *or* height (not both), in the format
    #                     '-pagewidth 3.i'  (3 inches width), or
    #                     '-pageheight 5.c' (5 cm high), for example
    #    papertype  -   letter | legal | A4 | A3
    #    crop       -    1 | 0  (1=>print only displayed region of canvas)
    #    xmargin    -    left/right margin, with units.  E.g.: 1.i
    #    ymargin    -    top/bottom margin,  as above
    #    position   -    One of: c nw n ne e se s sw w

    # Packing frames
    set col1frame [frame $window.row1frame -bd 0]
    set col2frame [frame $window.row2frame -bd 0]
    set outputwidgets [frame $window.outputwidgets -relief ridge -bd 4]
    set sizewidgets   [frame $window.sizewidgets   -relief ridge -bd 4]
    set actionwidgets [frame $window.actionwidgets -relief ridge -bd 4]

    # Use $printvar() for "local" global variables
    OMF_GetID $window id
    set printvar OMFpd$id
    global $printvar
    OMF_PrintDialogCopyIn $window $parrname ;# Set initial values

    # Filename
    OMF_MakeEntryWindow $window.file "Print to:" {} \
            ${printvar}(outname) 15 0 raised
    $window.file.value xview end   ;# Make end of string viewable
    bind $window.file.value <Key-Escape> \
	    "set ${printvar}(outname) \{$parr(outname)\}"

    # Orientation
    OMF_MakeRadioList $window.orient 1 0 "Orientation:" {} \
            ${printvar}(orient) {} Portrait Landscape

    # Paper type
    OMF_MakeRadioList $window.papertype 2 0 "Paper type" {} \
            ${printvar}(papertype) {} Letter Legal A4 A3

    # Position
    OMF_MakeRadioList $window.horpos 2 0 "Horizontal Position" {} \
            ${printvar}(horpos) {} {Left w} {Center ""} {Right e}
    OMF_MakeRadioList $window.verpos 2 0 "Vertical Position" {} \
            ${printvar}(verpos) {} {Top n} {Center ""} {Bottom s}

    # Units
    OMF_MakeRadioList $window.units 1 1 "Units:" {} \
            ${printvar}(units) {} {in i} {cm c}

    # Margins
    frame $window.marginframe -relief ridge -bd 2
    OMF_MakeEntryWindow $window.topmargin "Margin Top:" 12 \
            ${printvar}(topmargin) 6 0 flat
    OMF_MakeEntryWindow $window.leftmargin "Left:" 7 \
            ${printvar}(leftmargin) 6 0 flat
    pack $window.topmargin $window.leftmargin \
            -side left -in $window.marginframe

    # Print dimensions
    frame $window.dimframe -relief ridge -bd 2
    OMF_MakeEntryWindow $window.width "Print Width:" 12 \
            ${printvar}(width) 6 0 flat
    OMF_MakeEntryWindow $window.height "Height:" 7 \
            ${printvar}(height) 6 0 flat
    pack $window.width $window.height -side left -in $window.dimframe

    # Crop?
    frame $window.cropframe -relief ridge -bd 2
    checkbutton $window.crop -text "Crop to view"  -relief flat \
            -variable ${printvar}(crop)
    pack $window.crop -side left -in $window.cropframe -expand 1

    # Action buttons
    button $window.ok -text "OK" -width 8 -command \
            "OMF_ThreadSafe OMF_PrintDialogOK,$window \
	    \{OMF_PrintDialogOK $window $parrname\}"
    button $window.reset -text "Reset" -width 8 -command \
            "OMF_ThreadSafe OMF_PrintDialogCopyIn,$window \
	    \{OMF_PrintDialogCopyIn $window $parrname\}"
    button $window.cancel -text "Cancel" -width 8 -command \
            "OMF_ThreadSafe OMF_PrintDialogCancel,$window \
	    \{OMF_PrintDialogCancel $window $parrname\}"
    wm protocol $window WM_DELETE_WINDOW "$window.cancel invoke"

    # Pack
    pack $window.file $window.orient $window.papertype \
            $window.horpos $window.verpos \
            -side top -fill x -expand 1 -in $outputwidgets 
    pack $window.units $window.marginframe $window.dimframe \
            $window.cropframe \
            -side top -fill x -expand 1 -in $sizewidgets
    pack $window.ok $window.reset $window.cancel -side left -expand 1 \
            -in $actionwidgets

    pack $outputwidgets -side top -fill x -in $col1frame \
            -padx 5 -pady 5 -expand 1 -anchor n
    pack $sizewidgets   -side top -fill x -in $col2frame \
            -padx 5 -pady 5 -anchor n
    pack $actionwidgets -side top -in $col2frame -fill x \
             -padx 5 -pady 5 -expand 1 -anchor s
    pack $col1frame -side left -fill both -expand 1
    pack $col2frame -side left -fill y

    # Trace to keep printvar internally consistent
    trace variable $printvar w OMF_PrintDialogTrace

    # Trace to update printvar if viewport in . changes
    # This is a little tricky, because we need to send the
    # name $printvar through the trace (which is not tied
    # to $printvar, but to global variable UpdateViewCount.
    # There is probably a cleaner way to do this...  -mjd 97/2/11
    proc OMF_PrintDialogViewportTrace$id { name elt op } \
            "OMF_PrintDialogUpdateVariables $printvar"
    global UpdateViewCount
    trace variable UpdateViewCount w OMF_PrintDialogViewportTrace$id

    OMF_SetIcon $window
    focus $window.file.value
}
