/* FILE: vecfile.h                 -*-Mode: c++-*-
 *
 * Header file for vector field file I/O.
 *
 */

#include <stdio.h>
#include <tcl.h>
#include "machine.h"
#include "mesh.h"
#include "misc.h"
#include "template.h"
#include "vector.h"
#include "tclcmds.h"

class VectorFieldFileInput;  // Forward declaration

//////////////////////////////////////////////////////////////////////
// VectorFieldFileInputID
//
class VectorFieldFileInputID
{
  friend VectorFieldFileInput;
private:
  static const ClassDoc class_doc;
  String filetype;
  BOOL (*pIsType)(const char *filename,
		  FILE *input_file,
		  Tcl_Channel channel,
		  const char *channelName);
  VectorFieldFileInput* (*pMakeNew)(const char *new_filename,
				    FILE *new_infile,
				    Tcl_Channel new_channel,
				    const char *new_channelName);
  void Dup(const VectorFieldFileInputID &rhs); // Copy
public:
  // Constructor *with* registration (for global initialization)
  // (Set last field to NULL to avoid registration.)
  VectorFieldFileInputID(const char *new_filetype,
			 BOOL (*newpIsType)(const char *filename,
					    FILE *input_file,
					    Tcl_Channel channel,
					    const char *channelName),
			 VectorFieldFileInput *
			    (*newpNewReader)(const char *filename,
					     FILE *input_file,
					     Tcl_Channel channel,
					     const char *channelName),
			 BOOL (*pRegister)(const VectorFieldFileInputID &)
			 );

  // Copy constructor
  VectorFieldFileInputID(const VectorFieldFileInputID &copyobj);

  VectorFieldFileInputID(); // Default constructor; Needed by List<T>
  VectorFieldFileInputID& operator=(const VectorFieldFileInputID &rhs);

  ~VectorFieldFileInputID();
};

//////////////////////////////////////////////////////////////////////
// VectorFieldFileInput
//
class VectorFieldFileInput
{
  // NOTES ON CHILD CLASSES:
  //
  // Every child class should have a static member function defined
  // like this:
  //
  // static VectorFieldFileInput* MakeNew(const char *new_filename,
  //                                      FILE *new_infile,
  //                                      Tcl_Channel new_channel,
  //	                                  const char *new_channelName);
  //
  // which calls its "internal" constructor.  This function
  // is passed to VectorFieldFileInput::RegisterChild as part
  // of a VectorFieldFileInputID.
  //
  // All children classes should have a static member variable:
  //
  //   static VectorFieldFileInputID id;
  //
  // This initialization of this variable should call
  // VectorFieldFileInput::RegisterChild(id) so that they can be
  // called as necessary from VectorFieldFileInput::NewReader().
  //
  // The FILE *infile is closed in the base class destructor.

private:
  static const ClassDoc class_doc;
  static List<VectorFieldFileInputID> child_list;

  // The following 3 functions are left undefined on purpose
  VectorFieldFileInput(); // Default constructor
  VectorFieldFileInput(VectorFieldFileInput &copyobj);  // Copy constructor
  VectorFieldFileInput& operator=(VectorFieldFileInput &rhs); // Assignment op

protected:
  const String filename;
  String filetitle; // (Long) filetitle
  String filedesc;  // File description
  FILE *infile;
  Tcl_Channel channel;     // Channel registered with omfInterp
  const char *channelName; // (omfInterp is a global variable defined
  // in tclcmds.cc, which is a pointer to the current Tcl command
  // interpreter.)
  // NOTE: The file is "registered" with the Tcl interpreter inside
  // "NewReader", and so must be closed *only* by a corresponding
  // call to Tcl_UnregisterChannel(), which occurs in the
  // ~VectorFieldFileInput destructor, NOT IN ANY OF THE CHILD CLASSES!

  // Internal constructor
  VectorFieldFileInput(const char *new_filename,
		       FILE *new_infile,
		       Tcl_Channel new_channel,
		       const char *new_channelName);

  static BOOL RegisterChild(const VectorFieldFileInputID &id);

public:
  // Clients should call NewReader() to get a VFFI instance.  The
  // client is responsible for destroying (via delete) the returned
  // object, i.e., treat this as a "new" operator.
  static VectorFieldFileInput *NewReader(const char *new_filename);

  virtual ~VectorFieldFileInput();
  const char *GetFileName(void) { return (const char *)filename; }
  virtual const char *FileType(void)=0;
  virtual Mesh *NewMesh(void)=0;  // Derived classes should return
    // a *EmptyMesh on error (not NULL).  Either way, the client
    // is responsible for destroying (via delete) the returned mesh.
};

//////////////////////////////////////////////////////////////////////
// Omf File Format specs and utility routines
enum OmfDataStyle { omfi_ascii, omfi_bin4, omfi_bin8 };
class OmfFileFormatSpecs
{
private:
  static const ClassDoc class_doc;
  static const char* IDString;     // ID string at top of file
  static const REAL4 CheckValue4;  // 4-byte real data block check value
  static const REAL8 CheckValue8;  // 8-byte   "                     "
public:
  // Format of data in "Data" block: ascii, 4-byte binary, 8-byte binary

  static BOOL CheckIDString(const char *test);
  static BOOL CheckValue(REAL4 test) { return test==CheckValue4; }
  static BOOL CheckValue(REAL8 test) { return test==CheckValue8; }

  // The use of the CheckValue functions is encouraged, but
  // the actual values are needed for creating ovf files.
  static void GetCheckValue(REAL4 &check) { check=CheckValue4; }
  static void GetCheckValue(REAL8 &check) { check=CheckValue8; }

  // Binary I/O routines that perform byteswapping as needed.
  static INT4m ReadBinary(FILE *infile,REAL4 *buf,INT4m count);
  static INT4m ReadBinary(FILE *infile,REAL8 *buf,INT4m count);
  static INT4m WriteBinary(FILE *outfile,REAL4 *buf,INT4m count);
  static INT4m WriteBinary(FILE *outfile,REAL8 *buf,INT4m count);
  // NOTE: The write routines may modify buf during byteswapping.
  //   If this is unacceptable, then one should add additional
  //   routines that take a const buf, and copy buf into internal
  //   storage.

};

//////////////////////////////////////////////////////////////////////
// OmfSegmentHeader
//
enum OmfSegmentHeadMeshType { oshmt_rectangular, oshmt_irregular  };
struct OmfSegmentHeader
{
private:
  static const ClassDoc class_doc;
public:
  // HEADER VALUES
  String title; // Long file name or title goes here
  String desc;  // File description
  String meshunit; // Fundamental mesh measurement unit.  Treated as a label.
                   // Example: nm
  OmfSegmentHeadMeshType meshtype;

  REAL8 xbase,ybase,zbase; // (xbase,ybase,zbase) is the position, in
                           // 'meshunit', of the first point in the data
                           //  section (below).

  REAL8 xstepsize,ystepsize,zstepsize;  // Distance between adjacent
  //grid pts.: on the x-axis, 20 nm, etc.  The sign on these values
  // determines the grid orientation relative to (xbase,ybase,zbase).

  INT4 xnodes,ynodes,znodes; // Number of nodes along each axis, regular grids.
  INT4 pointcount;           // Number of points, for irregular grids.

  REAL8 xmin,ymin,zmin;  // Corner points defining mesh boundary.  The
  REAL8 xmax,ymax,zmax;  // bounding box is set to include both these
  /// points and the extent from ?base, ?stepsize and ?nodes.

  List<Vec3f> bdryline;  // Points defining a polyline, used to indicate
  /// boundary.  Optional; if missing, a bdry will be constructed from
  /// the ?min & ?max boundary points.

  String valueunit; // Fundamental field value unit, treated as a label
                    // Example: kA/m

  REAL8 valuemultiplier; // Multiply file values by this to get
                         // true value in 'valueunits'.

  REAL8 valuerangemaxmag,valuerangeminmag; // These are in file value
  // units, and are used as hints (or defaults) by postprocessing programs.
  //  The mmdisp program ignores any points with magnitude smaller than
  // valuerangeminmag, and uses valuerangemaxmag to scale inputs for display.
  // The use of these values is postprocessor dependent, and should be
  // interactively overrideable, but typically ValueRangeMaxMag should be
  // the maximum vector magnitude in the file data, and ValueRangeMinMag
  // should be a lower bound on non-zero file vector magnitudes.

  // BOOLEANS
  // The following boolean values indicate whether each of the above values
  // have been set (1=>set, 0=>not set).
  BOOL title_bool,desc_bool;
  BOOL meshunit_bool,xbase_bool,ybase_bool,zbase_bool;
  BOOL xstepsize_bool,ystepsize_bool,zstepsize_bool;
  BOOL xnodes_bool,ynodes_bool,znodes_bool;
  BOOL xmin_bool,ymin_bool,zmin_bool;
  BOOL xmax_bool,ymax_bool,zmax_bool;
  BOOL valueunit_bool,valuemultiplier_bool;
  BOOL valuerangemaxmag_bool,valuerangeminmag_bool;
  BOOL meshtype_bool,pointcount_bool,bdryline_bool;

  // Utility routines
  OmfSegmentHeader();
  void SetBools(BOOL newval);  // Sets all BOOL's to newval
  BOOL CompleteSet(); // Returns 1 iff a complete collection of
  /// the BOOL's are 1.
  void SetFieldValue(const String &field,const String &value);
};

//////////////////////////////////////////////////////////////////////
// OmfFileInput
//
class OmfFileInput : public VectorFieldFileInput
{
private:
  static const ClassDoc class_doc;
  static const char *file_type;
  static const VectorFieldFileInputID id;

  // File read helper routines
  INT4m ReadLine(String &buf);
  INT4m ParseContentLine(const String &buf,String &field,String &value);
  INT4m ReadTextVec3f(Vec3f &data);
  INT4m OmfFileInput::ReadVec3f(OmfDataStyle datastyle,Vec3f &data);

  // Internal constructor
  OmfFileInput(const char *new_filename,
	       FILE *new_infile,
	       Tcl_Channel new_channel,
	       const char *new_channelName);

  // The next 2 functions & the following "Default constructor"
  // are left undefined on purpose
  OmfFileInput(OmfFileInput &copyobj);  // Copy constructor
  OmfFileInput& operator=(OmfFileInput &rhs); // Assignment op

public:
  OmfFileInput();  // Default constructor
  // Once upon a time (sometime before Aug-1997), I had trouble
  // with the SGI CC compiler requiring at least 1 public
  // constructor, so I put the dummy "OmfFileInput();" constructor
  // declaration here (but with no definition).  This restriction
  // appears to have gone away with the newer IRIX 6.2 CC compiler,
  // but I'll leave this note here just in case.  Of course, actual
  // code is required to call "MakeNew" instead of a constructor.

  static VectorFieldFileInput* MakeNew(const char *new_filename,
				       FILE *new_infile,
				       Tcl_Channel new_channel,
				       const char *new_channelName);
  const char *FileType(void) { return file_type; }
  static BOOL IsType(const char *filename,
		     FILE *input_file,
		     Tcl_Channel channel,
		     const char *channelName);
  Mesh *NewMesh(void);
};

//////////////////////////////////////////////////////////////////////
// OmfFileOutput
//
class OmfFileOutput
{
private:
  static const ClassDoc class_doc;
public:
  OmfFileOutput() {}
  INT4m WriteMesh(Mesh *mesh,FILE* fptr,OmfDataStyle datastyle,
		  BOOL force_irreg=0,
		  const char *title=NULL,const char *desc=NULL);
  // Returns -1 if mesh type not supported,
  //          0 on success,
  //         >0 on other error.
  // If force_irreg is set, the output will be in the Omf irregular
  // grid file format (location & value) even if input data lies
  // on a rectangular grid.
  ~OmfFileOutput() {}
};

//////////////////////////////////////////////////////////////////////
// VioFileOutput
//
class VioFileOutput
{
private:
  static const ClassDoc class_doc;
public:
  VioFileOutput() {}
  INT4m WriteMesh(Mesh *mesh,FILE* fptr,const char *desc=NULL);
  // Returns -1 if mesh type not supported,
  //          0 on success,
  //         >0 on other error.
  ~VioFileOutput() {}
};

//////////////////////////////////////////////////////////////////////
// VioFileInput
//
class VioFileInput : public VectorFieldFileInput
{
private:
  static const ClassDoc class_doc;
  static const char *file_type;
  static const VectorFieldFileInputID id;

  // The following 3 functions are left undefined on purpose
  VioFileInput(); // Default constructor
  VioFileInput(OmfFileInput &copyobj);  // Copy constructor
  VioFileInput& operator=(OmfFileInput &rhs); // Assignment op

public:
  // Internal constructor
  VioFileInput(const char *new_filename,
	       FILE *new_infile,
	       Tcl_Channel new_channel,
	       const char *new_channelName);

  // Standard constructor
  static VectorFieldFileInput* MakeNew(const char *new_filename,
				       FILE *new_infile,
				       Tcl_Channel new_channel,
				       const char *new_channelName);
  const char *FileType(void) { return file_type; }
  static BOOL IsType(const char *filename,
		     FILE *input_file,
		     Tcl_Channel channel,
		     const char *channelName);
  Mesh *NewMesh(void); // Returns *EmptyMesh on error
  GridVec3f *NewRectGrid(void);  // Return NULL on error
};


//////////////////////////////////////////////////////////////////////
// Standard problem (or Simple) Vector Format (SVF)
//
class SvfFileInput : public VectorFieldFileInput
{
private:
  static const ClassDoc class_doc;
  static const char *file_type;
  static const VectorFieldFileInputID id;
  friend int OMF_SvfFileInputTA(ClientData,Tcl_Interp *interp,
					int argc,char** argv);

  // Instance data members
  GeneralMesh3f *mesh;

  // The following 3 functions are left undefined on purpose
  SvfFileInput(); // Default constructor
  SvfFileInput(OmfFileInput &copyobj);  // Copy constructor
  SvfFileInput& operator=(OmfFileInput &rhs); // Assignment op

  // Internal constructor
  SvfFileInput(const char *new_filename,
	       FILE *new_infile,
	       Tcl_Channel new_channel,
	       const char *new_channelName);

public:

  // Standard constructor
  static VectorFieldFileInput* MakeNew(const char *new_filename,
				       FILE *new_infile,
				       Tcl_Channel new_channel,
				       const char *new_channelName);
  const char *FileType(void) { return file_type; }
  static BOOL IsType(const char *filename,
		     FILE *input_file,
		     Tcl_Channel channel,
		     const char *channelName);
  Mesh *NewMesh(void); // Returns *EmptyMesh on error
  GridVec3f *NewRectGrid(void) { return NULL; } // Return NULL on error
  ~SvfFileInput();
};
