/* FILE: mifio.cc           -*-Mode: c++-*-
 *
 * I/O routines involving *.mif (Micromagnetic Input File) files.
 *
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "misc.h"
#include "maginit.h"
#include "mifio.h"
#include "zeeman.h"
#include "maganis.h"

/* End includes */

#define MifSlowPipeName  "SlowPipe"
#define MifFastPipeName  "FastPipe"
#define Mif2dSlabName    "2dSlab"
#define Mif3dSlabName    "3dSlab"
#define Mif3dchargeName  "3dCharge"
#define MifNoDemagName   "NoDemag"

void FieldRangeArray::Free(void)
 {
  if(RangeCount>0) {
    if(StartField!=(double *)NULL) delete[] StartField;
    if(StopField!=(double *)NULL)  delete[] StopField;
    if(StepCount!=(int *)NULL)     delete[] StepCount;
  }
  RangeCount=0;
  StartField=StopField=(double *)NULL;
  StepCount=(int *)NULL;
  RangesFilled=StepTotal=CurIndex=0;
}
int FieldRangeArray::AreSameField(double* field1,double* field2)
{ // Returns 1 if field1==field2, 0 otherwise
  int i;
  for(i=0;i<MIFVECDIM;i++)
    if(field1[i]!=field2[i]) return 0;
  return 1;
}

FieldRangeArray::FieldRangeArray(void)
{
  RangeCount=0;
  StartField=StopField=(double *)NULL;
  StepCount=(int *)NULL;
  RangesFilled=StepTotal=CurIndex=0;
}

void FieldRangeArray::MakeSpace(int newrangecount)
{
  if(newrangecount<1)
    PlainError(1,"Error in FieldRangeArray::MakeSpace: "
	     "Illegal rangecount (%d) request",newrangecount);
  if(newrangecount!=RangeCount) {
    Free(); // Release any previously allocated space
    RangeCount=newrangecount;
    if(!(StartField=new double[RangeCount*MIFVECDIM]) ||
       !(StopField=new double[RangeCount*MIFVECDIM])  ||
       !(StepCount=new int[RangeCount]))
      PlainError(1,"Error in FieldRangeArray::MakeSpace: %s",ErrNoMem);
  }
  RangesFilled=StepTotal=CurIndex=0;
  int i,j;
  for(i=0;i<RangeCount;i++) {
    StepCount[i]=0;
    for(j=0;j<MIFVECDIM;j++)
      StartField[MIFVECDIM*i+j]=StopField[MIFVECDIM*i+j]=0.0;
  }
}

int FieldRangeArray::AddRange(const double* start,const double* stop,
			      int steps)
{
  if(RangesFilled>=RangeCount) {
    PlainWarning("Warning in FieldRangeArray::AddRange: "
		"Attempt to add more ranges than specified; skipping");
    return RangesFilled;
  }
  if(start==(double *)NULL || stop==(double *)NULL || steps<0)
    PlainError(1,"Error in FieldRangeArray: %s",ErrBadParam);

  int i,onefieldflag;
  for(i=0,onefieldflag=1;i<MIFVECDIM;i++) {
    StartField[MIFVECDIM*RangesFilled+i]=start[i];
    StopField[MIFVECDIM*RangesFilled+i] =stop[i];
    if(start[i]!=stop[i]) onefieldflag=0;
  }
  if(onefieldflag) steps=0;  // Degenerate case: single field, no real "range"
  StepTotal+=(StepCount[RangesFilled]=steps);
  if(RangesFilled>0 &&
     !AreSameField(StopField+MIFVECDIM*(RangesFilled-1),
		   StartField+MIFVECDIM*RangesFilled))
    StepTotal+=1; // Step needed from last stop field to this start field

  return ++RangesFilled;
}

int FieldRangeArray::AddRange(const double* onefield)
{
  double start[MIFVECDIM],stop[MIFVECDIM];
  int i,steps;
  for(i=0;i<MIFVECDIM;i++) {
    if(onefield==(double *)NULL) start[i]=stop[i]=0.0;
    else                         start[i]=stop[i]=onefield[i];
  }
  steps=0;
  return AddRange(start,stop,steps);
}

int FieldRangeArray::AddRange(const char** param)
{
  double start[MIFVECDIM],stop[MIFVECDIM];
  int i,steps;
  for(i=0;i<MIFVECDIM;i++) {
    start[i]=Omf_atof(param[i]);
    stop[i]=Omf_atof(param[MIFVECDIM+i]);
  }
  steps=atoi(param[2*MIFVECDIM]);
  return AddRange(start,stop,steps);
}

int FieldRangeArray::ConsistencyCheck(void)
{
  if(RangeCount<1) return 0;  // Nothing to check
  int errcount=0;
  if(RangesFilled<RangeCount) {
    PlainWarning("Warning in FieldRangeArray::ConsistencyCheck: "
		 "Only %d ranges filled out of specified %d",
		 RangesFilled,RangeCount);
    errcount++;
  }
  if(StartField==(double *)NULL || StopField==(double *)NULL
     || StepCount==(int *)NULL) {
    PlainWarning("Warning in FieldRangeArray::ConsistencyCheck: "
		 "Field arrays not initialized");
    return ++errcount;
  }

  int i,tempstep=0;
  for(i=0;i<RangesFilled;i++) {
    tempstep+=StepCount[i];
    if(i>0 && !AreSameField(StopField+MIFVECDIM*(i-1),StartField+MIFVECDIM*i))
      tempstep++;
  }
  if(tempstep!=StepTotal) {
    PlainWarning("Warning in FieldRangeArray::ConsistencyCheck: "
		 "Incorrect value for StepTotal detected (%d instead of %d)",
		 StepTotal,tempstep);
    errcount++;
  }

  if(CurIndex<0 || CurIndex>StepTotal) {
    PlainWarning("Warning in FieldRangeArray::ConsistencyCheck: "
		 "CurIndex (%d) is out-of-bounds [0,%d]",
		 CurIndex,StepTotal);
    errcount++;
  }

  return errcount;
}

int FieldRangeArray::SetFieldIndex(int newindex)
{
  if(newindex<0 || newindex>StepTotal) newindex=StepTotal;
  return CurIndex=newindex;
}

void FieldRangeArray::StepIndex(int& errorcode)
{ // Sets errorcode=0 on success,
  //               -1 if already at last step
  errorcode=0;
  CurIndex++;
  if(CurIndex<0) CurIndex=0;
  else if (CurIndex>StepTotal) {
    CurIndex=StepTotal;
    MessageLocker::Append("Warning from FieldRangeArray: "
			  "Attempt to step past last step (%d)\n",
			  StepTotal);
    errorcode=-1;
  }
}

double *FieldRangeArray::Field(int index)
{
  static double field[MIFVECDIM];
  if(index<0 || index>StepTotal) index=StepTotal;
  int itemp,irange,extrastep,instep,rangesteps;
  extrastep=instep=rangesteps=0; // Safety
  for(itemp=irange=0;irange<RangeCount;irange++) {
    if(irange>0 && !AreSameField(StopField+MIFVECDIM*(irange-1),
				 StartField+MIFVECDIM*irange))
         extrastep=1;
    else extrastep=0;
    rangesteps=StepCount[irange];
    if(itemp+rangesteps+extrastep>=index) break;
    itemp+=rangesteps+extrastep;
  }
  if(irange>=RangeCount) { // Safety
    irange=RangeCount-1;
    instep=rangesteps=StepCount[irange];
  }
  else  instep=index-itemp-extrastep;

  int i; double c1,c2;
  if(rangesteps<1) { c1=1.0; c2=0.0; }
  else { c2=double(instep)/double(rangesteps);  c1=1.0-c2; }
  if(c1<0.0 || c1>1.0)
    PlainError(1,"Error in FieldRangeArray::Field: %s",ErrProgramming);
  for(i=0;i<MIFVECDIM;i++)
    field[i]=c1*StartField[MIFVECDIM*irange+i]
      +c2*StopField[MIFVECDIM*irange+i];
  return field;
}

double *FieldRangeArray::CurrentField(void)
{
  return Field(CurIndex);
}

FieldRangeArray::~FieldRangeArray(void) { Free(); }

//////////////////////////////////////////////////////////////////////

#define MIFBIGBUFSIZE 4096
#define MIFIDDELIMS ":="                 // Field ID delimiters
#define MIFDELIMS  " \t,;:\"\'`[]{}()="  // Line parameter delimiters
#define MIFCOMMENT '#'     // String denoting comment start

#define MIFMAXPARAMS 64    // Maximum number of parameters per input line

void MIF::Init(void)
{ // Set MIF member variables to dummy values
  int i;
  CleanUp();
  Ms=A=K1= -1.0;
  AnisType=mif_anis_notset;
  for(i=0;i<MIFVECDIM;i++) AnisDir1[i]=AnisDir2[i]=0.0;
  DemagType=mif_demag_notset;
  PartWidth=PartHeight=PartThickness=CellSize= -1.0;
  GeomType=mif_geom_notset;   GeomTypeParam="0.0";
  RandomizerSeed= -1;
  UserInteractionLevel= -1;
  UserReportCode= -1;
  InitialIncrement= -1.;
  MaxIterationCount=-1;
  ConvergeTorqueValue=-1.0;
  LogLevel=-1;
}

void MIF::CleanUp(void)
{
  if(AnisInitArgc>0 && AnisInitArgv!=(char **)NULL) {
    if(AnisInitArgv[0]!=NULL) delete[] AnisInitArgv[0];
    delete[] AnisInitArgv;
  }
  AnisInitArgc=0; AnisInitArgv=(char **)NULL;
  if(MagInitArgc>0 && MagInitArgv!=(char **)NULL) {
    if(MagInitArgv[0]!=NULL) delete[] MagInitArgv[0];
    delete[] MagInitArgv;
  }
  MagInitArgc=0; MagInitArgv=(char **)NULL;
  FieldRange.FreeSpace();
  InName.Free();
  OutBaseName.Free();
  if(AppliedField!=(Zeeman *)NULL) {
    delete AppliedField; AppliedField=(Zeeman *)NULL;
  }
}

MIF::MIF(void)
{
  AnisInitArgc=0; AnisInitArgv=(char **)NULL;
  MagInitArgc=0; MagInitArgv=(char **)NULL;
  AppliedField=(Zeeman *)NULL;
  SetDefaults();
}

MIF::~MIF(void)
{
  CleanUp();
}

void MIF::SetDefaults(void)
{ // Set some member variables to default values
  int i;
  Init(); // Clear and set variables to dummy values
  for(i=0;i<MIFVECDIM;i++) AnisDir1[i]=AnisDir2[i]=0.0;
  AnisDir1[0]=AnisDir2[1]=1.0;
  AnisInitArgc=1;
  AnisInitArgv=new char*[AnisInitArgc];
  AnisInitArgv[0]=new char[strlen("constant")+1];
  strcpy(AnisInitArgv[0],"constant");
  DemagType=mif_fastpipe;
  GeomType=mif_rectangle;   
  MagInitArgc=1;
  MagInitArgv=new char*[MagInitArgc];
  MagInitArgv[0]=(char *)NULL;
  FieldRange.MakeSpace(1);
  FieldRange.AddRange(); // Zero field, no steps.
  RandomizerSeed=0;
  UserInteractionLevel=0;
  UserReportCode=0;
  LogLevel=0;
  MaxIterationCount=-1; // No limit
  ConvergeTorqueValue=1e-4;
  InitialIncrement= 1.0*PI/180.;  // 1 degree.
  const char *cpptr[1]; cpptr[0]="uniform";
  AppliedField=Zeeman::MakeZeeman(1,cpptr);
}

// MIF::ReadLine reads and tokenizes 1 logical line from infile.
// A non-zero return value indicates error; 1 => end-of-file
#define MRLsuccess  0  // Successful read
#define MRLeof      1  // Proper end-of-file
#define MRLbadeof   2  // Improper eof
#define MRLferror   3  // Other file error
#define MRLbufof    4  // Input buffer overflow
#define MRLparamof  5  // Parameter space overflow
#define MRLbadlist  6  // Parameters don't form proper Tcl list

int MIF::ReadLine(FILE* infile,int max_strc,
		  int &strc,String* strv)
{
  static char buf[MIFBIGBUFSIZE];
  static String line(MIFBIGBUFSIZE);

  strc=0;

  // Read one logical line
  buf[sizeof(buf)-1]=buf[sizeof(buf)-2]='\0';  // Safety; Shouldn't be needed.
  line="";
  int read_length,line_done,eof_flag=0;
  do {
    line_done=0;
    buf[0]='\0'; // Safety
    if(fgets(buf,sizeof(buf)-1,infile)==NULL) {
      eof_flag=1; break;   // Assume error is EOF  
    }
    read_length=strlen(buf);
    if(buf[read_length-1]=='\n') {
      read_length--;
      if(buf[read_length-1]=='\\') read_length--;
      else                         line_done=1;
    }
    line.Append(buf,read_length);
  } while(!line_done);
  line.TrimLeft();
  if(eof_flag && line.Length()==0) return MRLeof;

  if(line[0]=='\0' || line[0]==MIFCOMMENT)
    return MRLsuccess; // Comment or blank line

  // Pull out and trim field identifier
  if(max_strc<1) return MRLparamof;
  int line_len=line.Length();
  char *workbuf=new char[line_len+2]; // Be sure to delete[] before rtn!
  workbuf[line_len+1]='\0'; // Safety
  strcpy(workbuf,line.GetStr());
  int offset=strcspn(workbuf,MIFIDDELIMS);
  if(offset+1>=line_len) {
    // Field ID terminator not found; ASSUME this is a comment line
    delete[] workbuf; return MRLsuccess;
  }
  workbuf[offset]='\0';
  strv[0].Dup(workbuf);
  strv[0].Trim();
  strc=1;
  offset=offset+1+strspn(workbuf+offset+1,MIFIDDELIMS);
  if(offset>=line_len) {
    // Nothing after end of field ID terminator! Bad line...
    delete[] workbuf; return MRLbadlist;
  }

  // Assume any backslashes in remainder are filename path separators,
  // and translate them into forward slashes.  This is necessary because
  // Tcl_SplitList (used to parse the arguments) does backslash
  // substitution.
  for(char *cp=workbuf+offset;*cp!='\0';cp++) {
    if(*cp=='\\') *cp='/';
  }

  // Strip comments and break remainder into list elements
  // Comments are detected by truncating the line at each
  // comment character, starting from the left, and testing
  // to see if we have a proper Tcl list.  If a comment char
  // is buried inside braces or quotes, then the string to
  // the left of the comment character will not be a proper
  // list, in which case we reinstate the character and
  // retry at the next comment character (if any).
  char *cmtptr;
  int paramc;  char **paramv;
  cmtptr=workbuf+offset;
  while(1) {
    if((cmtptr=strchr(cmtptr,MIFCOMMENT))!=NULL) 
      *cmtptr='\0';  // Truncate comment
    // In the following, replace NULL with omfInterp if you want
    // to get back an error message.
    if(Tcl_SplitList(NULL,workbuf+offset,&paramc,&paramv)==TCL_OK) {
      // Proper list structure => success
      break;
    } else {
      // Improper list structure
      if(cmtptr!=NULL) { // Bad comment truncation
	*(cmtptr++)=MIFCOMMENT;
      } else {           // Bad entry
	delete[] workbuf;
	return MRLbadlist;
      }
    }
  }

  // Copy parameter strings into strv
  int error_code=MRLsuccess;
  for(int i=0;i<paramc;i++) {
    if(strc>=max_strc) { error_code=MRLparamof; break; } // Overflow
    strv[strc++].Dup(paramv[i]);
  }

  delete[] workbuf;
  omf_Tcl_Free((char *)paramv);
  return error_code;
}

// Fixup AnisDir1 and AnisDir2 as necessary
#define MIFNAME "MIF::AdjustAnisDirs: "
void MIF::AdjustAnisDirs(void)
{
  int i; double temp;
  // Make AnisDir1 a unit vector
  for(i=0,temp=0.0;i<MIFVECDIM;i++) temp+=SQ(AnisDir1[i]);
  if(temp<SQRT_EPSILON)
    PlainError(1,"Error in " MIFNAME "AnisDir1 has zero length");
  if(temp!=1.0) for(i=0,temp=sqrt(temp);i<MIFVECDIM;i++) AnisDir1[i]/=temp;

  // If AnisType is cubic, then AnisDir2 should be unit vector perpendicular
  // to AnisDir1
  if(AnisType==mif_cubic) {
    for(i=0,temp=0.0;i<MIFVECDIM;i++) temp+=AnisDir1[i]*AnisDir2[i];
    if(temp!=0.0) { // Eliminate AnisDir1 component from AnisDir2
      for(i=0;i<MIFVECDIM;i++) AnisDir2[i]-=temp*AnisDir1[i];
    }
    // Make AnisDir2 a unit vector
    for(i=0,temp=0.0;i<MIFVECDIM;i++) temp+=SQ(AnisDir2[i]);
    if(temp<SQRT_EPSILON) PlainError(1,"Error in " MIFNAME
          "Component of AnisDir2 perpendicular to AnisDir1 has zero length");
    if(temp!=1.0) for(i=0,temp=sqrt(temp);i<MIFVECDIM;i++) AnisDir2[i]/=temp;
  }
}
#undef MIFNAME

#define MIFNAME "MIF::ConsistencyCheck: "
#define MIFERRF(x) \
   PlainWarning("Warning in " MIFNAME "Illegal value for %s: %f",#x,x),\
   errcount++
#define MIFERRD(x) \
   PlainWarning("Warning in " MIFNAME "Illegal value for %s: %d",#x,x),\
   errcount++
#define MIFERRE(v,s) \
   PlainWarning("Warning in " MIFNAME "Illegal value for %s: %s",#v,#s),\
   errcount++

// Routine to perform some consistency checks on member variables
// Return value is the number of errors detected
int MIF::ConsistencyCheck(void)
{
  int i,errcount; double temp;
  errcount=0;
  // Material parameters
  if(Ms<0.0) MIFERRF(Ms);
  if(A<0.0)  MIFERRF(A);
  if(AnisType==mif_anis_notset) MIFERRE(AnisType,mif_anis_notset);
  for(i=0,temp=0.0;i<MIFVECDIM;i++) temp+=SQ(AnisDir1[i]);
  if(fabs(temp-1.0)>EPSILON)
    PlainWarning("Warning in " MIFNAME "AnisDir1 has length %f != 1",
		 sqrt(temp)),errcount++;
  if(AnisType==mif_cubic) { // Check for illegal AnisDir2
    for(i=0,temp=0.0;i<MIFVECDIM;i++) temp+=SQ(AnisDir2[i]);
    if(fabs(temp-1.0)>EPSILON)
      PlainWarning("Warning in " MIFNAME "AnisDir2 has length %f != 1",
		   sqrt(temp)),errcount++;
    for(i=0,temp=0.0;i<MIFVECDIM;i++) temp+=AnisDir1[i]*AnisDir2[i];
    if(fabs(temp)>EPSILON)
      PlainWarning("Warning in " MIFNAME "AnisDir2 not perpendicular to"
		   " AnisDir1; Inner product =%f",temp),errcount++;
  }
  if(AnisInitList.GetAnisInitFunc(AnisInitArgv[0])==NULL)
    PlainWarning("Warning in " MIFNAME "Unknown AnisInitFunc: ->%s<-",
		 AnisInitArgv[0]),errcount++;

  // Demag specification
  if(DemagType==mif_demag_notset) MIFERRE(DemagType,mif_demag_notset);

  // Part geometry
  if(PartWidth<=0.0)     MIFERRF(PartWidth);
  if(PartHeight<=0.0)    MIFERRF(PartHeight);
  if((DemagType==mif_2dslab || DemagType==mif_3dslab)
     && PartThickness<=0.0) MIFERRF(PartThickness);
  if(CellSize<=0.0)      MIFERRF(CellSize);
  if(CellSize>PartWidth) {
    PlainWarning("Warning in " MIFNAME "CellSize=%g > PartWidth=%g",
		 CellSize,PartWidth);
    errcount++;
  }
  if(CellSize>PartHeight) {
    PlainWarning("Warning in " MIFNAME "CellSize=%g > PartHeight=%g",
		 CellSize,PartHeight);
    errcount++;
  }
  if(GeomType==mif_geom_notset)   MIFERRE(GeomType,mif_geom_notset);


  // Magnetization initialization
  if(MagInit::NameCheck(MagInitArgv[0])==NULL)
    PlainWarning("Warning in " MIFNAME
		 "Unknown magnetization initialization routine %s",
		 MagInitArgv[0]),errcount++;
  else {
    if(MagInit::GetParamCount(MagInitArgv[0])!=MagInitArgc-1)
      PlainWarning("Warning in " MIFNAME
		  "Wrong number of parameters to mag init routine %s,"
		  " %d instead of %d",MagInitArgv[0],MagInitArgc-1,
		  MagInit::GetParamCount(MagInitArgv[0])),errcount++;
  }

  // Experiment parameters
  errcount+=FieldRange.ConsistencyCheck();
  if(AppliedField==(Zeeman *)NULL) {
      PlainWarning("Warning in " MIFNAME "External field type not set");
      errcount++;
  }

  // Output specification
  if(OutBaseName.Length()<1) {
    PlainWarning("Warning in " MIFNAME "Output filename not set");
    errcount++;
  }
  if(LogLevel<0) {
    PlainWarning("Warning in " MIFNAME "Invalid Log level %d<0",LogLevel);
    errcount++;
  }

  // Miscellaneous
  if(InitialIncrement<=0.0)
    PlainWarning("Warning in " MIFNAME "Initial increment non-positive (%f)",
		InitialIncrement),errcount++;
  if(UserInteractionLevel<0 || UserInteractionLevel>100)
    MIFERRD(UserInteractionLevel);

  return errcount;
}
#undef MIFNAME
#undef MIFERRF
#undef MIFERRD
#undef MIFERRE


#define MIFNAME "MIF::Read: "
#define MIFCOUNT "Incorrect number of parameters (%d) for entry \"%s\""
void MIF::Read(char* inname,int &errorcode)
{ // Returns "".  Argument errorcode == 0 if no error,
  //                                  > 0 on fatal error.
  //                                  < 0 on non-fatal error(s).
  //  Use MessageLocker::GetMessage(String &) to retrieve
  //  error and warning messages (if errorcode!=0).

  int i,paramcount,readcode;
  String param[MIFMAXPARAMS];

  errorcode=1;     // Cleared after successful completion
  int warncount=0; // Number of non-fatal warnings

  SetDefaults();
  InName=inname;

  FILE* infile;
  if(!(infile=Omf_FOpen(InName.GetStr(),"r"))) {
    // Add a .mif extension and retry
    InName.Append(".mif");
    if(!(infile=Omf_FOpen(InName.GetStr(),"r"))) {
      MessageLocker::Append("Error in " MIFNAME "Unable to open file %s\n",
			    inname);
      return;
    }
  }

  do {
    // Read input line
    if((readcode=ReadLine(infile,sizeof(param)/sizeof(param[0]),
			  paramcount,param))!=MRLsuccess) break;

    if(paramcount<1) continue; // Empty line

    param[0].StripWhite(); param[0].ToLower();

    // MATERIAL PARAMETERS
    if(strcmp(param[0],"ms")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      Ms=Omf_atof(param[1]);
    }
    else if(strcmp(param[0],"a")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      A=Omf_atof(param[1]);
    }
    else if(strcmp(param[0],"k1")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      K1=Omf_atof(param[1]);
    }
    else if(strcmp(param[0],"anisotropytype")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      param[1].Trim(); param[1].CollapseStr(); param[1].ToLower();
      if(strcmp(param[1],"uniaxial")==0)   AnisType=mif_uniaxial;
      else if(strcmp(param[1],"cubic")==0) AnisType=mif_cubic;
      else {
	MessageLocker::Append("Warning in " MIFNAME
			      "Unknown AnisType: %s\n",param[1].GetStr());
	warncount++;
      }
    }
    else if(strcmp(param[0],"anisotropydir1")==0) {
      if(paramcount!=1+MIFVECDIM) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      for(i=0;i<MIFVECDIM;i++) AnisDir1[i]=Omf_atof(param[1+i]);
    }
    else if(strcmp(param[0],"anisotropydir2")==0) {
      if(paramcount!=1+MIFVECDIM) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      for(i=0;i<MIFVECDIM;i++) AnisDir2[i]=Omf_atof(param[1+i]);
    }

    // Anisotropy initialization function
    else if(strcmp(param[0],"anisotropyinitialization")==0) {
      if(paramcount<2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      if(AnisInitArgc>0 && AnisInitArgv!=NULL) {
	if(AnisInitArgv[0]!=NULL) delete[] AnisInitArgv[0];
	delete[] AnisInitArgv;
      }
      AnisInitArgv=new char*[AnisInitArgc=paramcount-1];
      int totalsize=0;
      for(i=0;i<AnisInitArgc;i++) totalsize+=(param[1+i].Length()+1);
      AnisInitArgv[0]=new char[totalsize+1];  // "+1" for safety
      for(i=0;i<AnisInitArgc;i++) {
	strcpy(AnisInitArgv[i],param[1+i]);
	if(i+1<AnisInitArgc)
	  AnisInitArgv[i+1]=
	    (char *)strchr((const char *)AnisInitArgv[i],'\0')+1;
      }
    }

    // DEMAG SPECIFICATION
    else if(strcmp(param[0],"demagtype")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      param[1].Trim(); param[1].CollapseStr(); param[1].ToLower();
      if(strcmp(param[1],"slowpipe")==0)      DemagType=mif_slowpipe;
      else if(strcmp(param[1],"fastpipe")==0) DemagType=mif_fastpipe;
      else if(strcmp(param[1],"2dslab")==0)   DemagType=mif_2dslab;
      else if(strcmp(param[1],"3dslab")==0)   DemagType=mif_3dslab;
      else if(strcmp(param[1],"3dcharge")==0) DemagType=mif_3dcharge;
      else if(strcmp(param[1],"none")==0)     DemagType=mif_nodemag;
      else {
	MessageLocker::Append("Warning in " MIFNAME "Unknown DemagType: %s\n",
			      param[1].GetStr());
	warncount++;
      }
    }
    // PART GEOMETRY
    else if(strcmp(param[0],"partwidth")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      PartWidth=Omf_atof(param[1]);
    }
    else if(strcmp(param[0],"partheight")==0) {
      if(paramcount!=2)
	PlainError(1,"Error in " MIFNAME MIFCOUNT,paramcount-1,
		   param[0].GetStr());
      PartHeight=Omf_atof(param[1]);
    }
    else if(strcmp(param[0],"partthickness")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      PartThickness=Omf_atof(param[1]);
    }
    else if(strcmp(param[0],"cellsize")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      CellSize=Omf_atof(param[1]);
    }
    else if(strcmp(param[0],"partshape")==0) {
      if(paramcount<2 || paramcount>3) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      param[1].Trim(); param[1].CollapseStr(); param[1].ToLower();
      if(strcmp(param[1],"rectangle")==0)    GeomType=mif_rectangle;
      else if(strcmp(param[1],"ellipse")==0) GeomType=mif_ellipse;
      else if(strcmp(param[1],"oval")==0)    GeomType=mif_oval;
      else if(strcmp(param[1],"mask")==0)    GeomType=mif_mask;
      else {
	MessageLocker::Append("Warning in " MIFNAME "Unknown GeomType: %s\n",
			      param[1].GetStr());
	warncount++;
      }
      if(paramcount==3) GeomTypeParam=param[2];
      else              GeomTypeParam="0.0";
    }
    // INITIAL MAGNETIZATION
    else if(strcmp(param[0],"initmag")==0) {
      if(paramcount<2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      if(MagInitArgc>0 && MagInitArgv!=NULL) {
	if(MagInitArgv[0]!=NULL) delete[] MagInitArgv[0];
	delete[] MagInitArgv;
      }
      MagInitArgv=new char*[MagInitArgc=paramcount-1];
      int totalsize=0;
      for(i=0;i<MagInitArgc;i++) totalsize+=(param[1+i].Length()+1);
      MagInitArgv[0]=new char[totalsize+1];  // "+1" is safety
      for(i=0;i<MagInitArgc;i++) {
	strcpy(MagInitArgv[i],param[1+i]);
	if(i+1<MagInitArgc)
	  MagInitArgv[i+1]=
	    (char *)strchr((const char *)MagInitArgv[i],'\0')+1;
      }
    }
    // EXPERIMENT PARAMETERS
    else if(strcmp(param[0],"fieldrangecount")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      FieldRange.MakeSpace(atoi(param[1]));
    }
    else if(strcmp(param[0],"fieldrange")==0) {
      if(paramcount!=2*MIFVECDIM+2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      const char **argp=new const char*[paramcount-1];
      for(int k=0;k<paramcount-1;k++) argp[k]=param[k+1].GetStr();
      FieldRange.AddRange(argp);
      delete[] argp;
    }
    else if(strcmp(param[0],"fieldtype")==0) {
      if(paramcount<2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      if(AppliedField!=(Zeeman *)NULL) delete AppliedField;
      const char **argp=new const char*[paramcount-1];
      for(int k=0;k<paramcount-1;k++) argp[k]=param[k+1].GetStr();
      AppliedField=Zeeman::MakeZeeman(paramcount-1,argp);
      delete[] argp;
    }

    // OUTPUT SPECIFICATION
    else if(strcmp(param[0],"baseoutputfilename")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      OutBaseName=param[1];
    }
    else if(strcmp(param[0],"loglevel")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      LogLevel=atoi(param[1]);
    }
    // MISCELLANEOUS
    else if(strcmp(param[0],"randomizerseed")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      RandomizerSeed=atoi(param[1]);
    }
    else if(strcmp(param[0],"initialincrement")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      InitialIncrement=Omf_atof(param[1])*PI/180.;
    }
    else if(strcmp(param[0],"userinteractionlevel")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      UserInteractionLevel=atoi(param[1]);
    }
    else if(strcmp(param[0],"userreportcode")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      UserReportCode=atoi(param[1]);
    }
    else if(strcmp(param[0],"maxiterationcount")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      MaxIterationCount=atoi(param[1]);
    }
    else if(strcmp(param[0],"convergetorquevalue")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0].GetStr());
	return;
      }
      ConvergeTorqueValue=Omf_atof(param[1]);
    }
    // UNKNOWN
    else {
      MessageLocker::Append("Warning in " MIFNAME "Unknown line tag: %s\n",
			    param[0].GetStr());
      warncount++;
    }
  } while(readcode==0); // End of input read loop
  fclose(infile);

  // Check for input errors
  if(readcode!=MRLeof) {
    if(readcode==MRLbadeof)
      MessageLocker::Append("Error in " MIFNAME "Improper eof\n");
    else if(readcode==MRLferror)
      MessageLocker::Append("Error in " MIFNAME "File error\n");
    else if(readcode==MRLbufof)
      MessageLocker::Append("Error in " MIFNAME "Input buffer overflow\n");
    else if(readcode==MRLparamof)
      MessageLocker::Append("Error in " MIFNAME "Parameter space overflow\n");
    else 
      MessageLocker::Append("Error in " MIFNAME "Unknown file input error\n");
    return;
  }

  // Adjust AnisDir1 and AnisDir2
  AdjustAnisDirs();

  // Check for valid specification
  int errcount;
  if((errcount=ConsistencyCheck())!=0) {
    MessageLocker::Append("Error in " MIFNAME 
			  "%d error%s found in input file %s\n",
			  errcount,(errcount==1?"":"s"),inname);
    return;
  }

  // If we get here, then there are no fatal errors.
  errorcode = -warncount;  // Possibly some warnings, though.

  return;
}
#undef MIFNAME
#undef MIFCOUNT

const char* MIF::GetDemagName(void)
{
  const char* cptr;
  switch(DemagType)
    {
    case mif_slowpipe:  cptr=MifSlowPipeName; break;
    case mif_fastpipe:  cptr=MifFastPipeName; break;
    case mif_2dslab:    cptr=Mif2dSlabName;   break;
    case mif_3dslab:    cptr=Mif3dSlabName;   break;
    case mif_3dcharge:  cptr=Mif3dchargeName;   break;
    case mif_nodemag:   cptr=MifNoDemagName;  break;
    default:            cptr=MifFastPipeName; break;
    }
  return cptr;
}

const double* MIF::GetAnisDir(int i)
{
  if(i==1) return AnisDir1;
  if(i==2) return AnisDir2;
  return (const double*)NULL;
}

void MIF::DumpPartialInfo(FILE* fptr)
{
  int i;
  fprintf(fptr,"# MIF input file: %s\n",InName.GetStr());
  fprintf(fptr,"#  Material information---\n");
  fprintf(fptr,"#   Ms=%.5e  A=%.5e  K1=%.5e\n",Ms,A,K1);
  fprintf(fptr,"#   AnisType= %s\n",
	  (AnisType==mif_uniaxial ? "Uniaxial" :
	   (AnisType==mif_cubic ? "Cubic" : "Not set")));
  fprintf(fptr,"#   AnisDir1=(%6.3f,%6.3f,%6.3f)\n",
	  AnisDir1[0],AnisDir1[1],AnisDir1[2]);
  if(AnisType==mif_cubic)
    fprintf(fptr,"#   AnisDir2=(%6.3f,%6.3f,%6.3f)\n",
	    AnisDir2[0],AnisDir2[1],AnisDir2[2]);
  fprintf(fptr,"#   AnisInit=");
  for(i=0;i<AnisInitArgc;i++) fprintf(fptr," %s",AnisInitArgv[i]);
  fprintf(fptr,"\n");
  fprintf(fptr,"#  Demag type: ");
  switch(DemagType)
    {
    case mif_slowpipe:     fprintf(fptr,"Slow Pipe\n"); break;
    case mif_fastpipe:     fprintf(fptr,"Fast Pipe\n"); break;
    case mif_2dslab:       fprintf(fptr,"2D Slab\n"); break;
    case mif_3dslab:       fprintf(fptr,"3D Slab\n"); break;
    case mif_3dcharge:       fprintf(fptr,"3D Charge\n"); break;
    case mif_nodemag:      fprintf(fptr,"No demag\n"); break;
    case mif_demag_notset: fprintf(fptr,"Not set\n"); break;
    default:           fprintf(fptr,"ERROR: UNKNOWN DEMAG TYPE\n"); break;
    }
  fprintf(fptr,"#  Geometry information---\n");
  fprintf(fptr,"#   Part width:     %.4e m\n",PartWidth);
  fprintf(fptr,"#   Part height:    %.4e m\n",PartHeight);
  fprintf(fptr,"#   Part thickness: %.4e m\n",PartThickness);
  fprintf(fptr,"#   Cell size:      %.4e m\n",CellSize);
  fprintf(fptr,"#   Part shape:     %s %s\n",
	  (GeomType==mif_rectangle ? "rectangle" :
	   (GeomType==mif_ellipse ? "ellipse" :
	   (GeomType==mif_mask ? "Mask" : 
	    (GeomType==mif_oval ? "oval" : "not set")))),GeomTypeParam.GetStr());
  fprintf(fptr,"#  Initial Magnetization:");
  for(i=0;i<MagInitArgc;i++) fprintf(fptr," %s",MagInitArgv[i]);
  fprintf(fptr,"\n");
  fprintf(fptr,"#  Randomizer seed: %d\n",RandomizerSeed);
  fprintf(fptr,"#  Output base name: %s\n",OutBaseName.GetStr());
  fflush(fptr);
}

