/* FILE: vecio.cc             -*-Mode: c++-*-
 *
 * This file contains class definitions for vector input/output routines.
 *
 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include "constants.h"
#include "misc.h"
#include "vecio.h"

/* End includes */

#ifdef ALL_CHECKS
#define CODE_CHECKS
#define INFO_LEVEL 10
#endif

#undef ORIGCOLORS
#define FULLINTENSITY 65535
#define BACKBRIGHT    0.45

// Gnu gcc wants these definitions to be inside the class declaration,
// but some other C++ compilers only accept them outside...
// See also the file vecio.h
#ifndef __GNUC__
const int    FileVector::ArrowVertexCount=8;
const double FileVector::ArrowLineWidth=0.05;
const double FileVector::ArrowLength=0.90;
const double FileVector::ArrowHeadRatio=0.375;
const double FileVector::VertMagCrossover=0.1;
const int    FileVector::CrossVertexCount=12;
const int    FileVector::MaxColorValue=255;
const double FileVector::BitmapCircleRadius=0.25;
const double FileVector::BitmapDotRadius=0.05;
const double FileVector::BitmapLineWidth=0.08;  
const double FileVector::BitmapArrowHeadRatio=0.45;
const double FileVector::BitmapArrowMinPixels=3.;
const double FileVector::PageWidth=(8.5*72);
const double FileVector::PageHeight=(11.*72);
const double FileVector::PageLeftMargin=(1.0*72);
const double FileVector::PageTopMargin= (1.0*72);
const int    FileVector::MinGrey=0;
const int    FileVector::MaxGrey=255;
#endif // !__GNUC__

const char   VecFileMagic02[8] = "VecFil\x02";
const double VecFileEndianTest = 1.234;
// The bit pattern for 1.234 on a little endian machine (e.g.,
// Intel's x86 series) should be 58 39 B4 C8 76 BE F3 3F, while
// on a big endian machine it is 3F F3 BE 76 C8 B4 39 58.  Examples
// of big endian machines include ones using the MIPS R4000 and R8000
// series chips, for example SGI's.

static void HSVtoRGB(double H,double S,double V,
		     double& R,double& G,double& B)
// Converts Hue-Saturation-Value to Red-Green-Blue.  All imports and
// exports in the range [0,1].  This routine based on email from John
// Hagedorn (john.hagedorn@nist.gov), NIST x4339, on 3-Oct-1995.
{
  H-=floor(H);  // Take hue input mod 1
  if(S>1.) S=1.; else if (S<0.) S=0.;
  if(V>1.) V=1.; else if (V<0.) V=0.;
  if(V==0) { R=G=B=0; return; }
  if(S==0) { R=G=B=V; return; }
  int sextant;  double fraction,p,q,t;
  sextant=(int)floor(H*6); fraction=H*6-sextant;
  p=V*(1.-S);  q=V*(1.-S*fraction);  t=V*(1.-S*(1.-fraction));
  switch(sextant) {
  default: // Safety
  case 0: // Red     -> Yellow
    R=V; G=t; B=p; break;
  case 1: // Yellow  -> Green
    R=q; G=V; B=p; break;
  case 2: // Green   -> Cyan
    R=p; G=V; B=t; break;
  case 3: // Cyan    -> Blue
    R=p; G=q; B=V; break;
  case 4: // Blue    -> Magenta
    R=t; G=p; B=V; break;
  case 5: // Magenta -> Red
    R=V; G=p; B=q; break;
  }
}

Polygon::Polygon(void)
{
  count=0;
  vertex=(TwoPoint *)NULL;
}

Polygon::Polygon(int vertex_count)
{
  if(vertex_count<2)
    PlainError(1,"Error in Polygon::Polygon(int): %s",ErrBadParam);
  count=vertex_count;
  if(!(vertex=new TwoPoint[count]))
    PlainError(1,"Error in Polygon::Polygon(int)L %s",ErrNoMem);
}

Polygon::Polygon(Polygon& poly)
{
  Polygon(poly.count);
  for(int i=0;i<count;i++) vertex[i]=poly.vertex[i];
}

Polygon::~Polygon()
{
  if(vertex!=(TwoPoint *)NULL) delete vertex;
}

void Polygon::Resize(int new_vertex_count)
{
  if(vertex!=(TwoPoint *)NULL) delete vertex;
  if(new_vertex_count<2)
    PlainError(1,"Error in Polygon::Resize(int): %s",ErrBadParam);
  count=new_vertex_count;
  if(!(vertex=new TwoPoint[count]))
    PlainError(1,"Error in Polygon::Polygon(int)L %s",ErrNoMem);
}

TwoPoint& Polygon::operator[](int index)
{
  if(index<0 || index>=count)
    PlainError(1,"Error in Polygon::operator[](int): %s",ErrBadParam);
  return vertex[index];
}

Polygon& Polygon::operator=(Polygon& poly)
{
  if(poly.count!=count)
    PlainError(1,"Error in Polygon::operator=(int): "
	       "Polygons have different vertex counts");
  for(int i=0;i<count;i++) vertex[i]=poly.vertex[i];
  return *this;
}

// Polygon::CrossSegment returns 1 if the ray from point "Q" in
// direction (1,eps) (where eps is an arbitrarily small positive
// value) crosses the line segment between points P1 and P2.
// Otherwise the return value is 0.  See page 11-12 (7-Sept-95)
// of MJD's micromagnetic NIST lab notebook.  This routine is used
// by  Polygon::IsInside().
int Polygon::CrossSegment(TwoPoint& Q,TwoPoint& P1,TwoPoint& P2)
{
  double delta;  // Discriminant for 2x2 inversion matrix
  delta=P2.y-P1.y;
  if(fabs(delta)<EPSILON) return 0; // Ray and lines are parallel.
  // NOTE: Do _not_ confuse EPSILON with eps is preceding discussion.

  // Check to see if intersection is between P1 and P2
  if(delta>0) {
    if(Q.y<P1.y || Q.y>P2.y) return 0;   // Intersection is outside segment
    if(Q.y==P1.y && Q.x>=P1.x) return 0; // Intersection is to left of P1
    if(Q.y==P2.y && Q.x<=P1.x) return 0; // Intersection it to right of P2
  }
  else {
    if(Q.y<P2.y || Q.y>P1.y) return 0;   // Intersection is outside segment
    if(Q.y==P2.y && Q.x>=P1.x) return 0; // Intersection is to left of P2
    if(Q.y==P1.y && Q.x<=P1.x) return 0; // Intersection it to right of P1
  }

  // Check to see if interection is on positive side of ray
  double s;
  s=(P1.x-Q.x)*(delta)+(Q.y-P1.y)*(P2.x-P1.x);
  if((s<0 && delta>0) || (s>0 && delta<0)) return 0;  // Negative side

  // Else we have a valid intersection
  return 1;
}

// Polygon::IsInside returns 1 if "pt" is inside the polygon,
// 0 otherwise.  The technique is to draw a ray from "pt" in
// direction (1,eps) (where eps is an arbitrarily small positive
// value) and count how many polygon edges get crossed.  An even
// count occurs for points outside the polygon, odd for inside.
// The eps shift is used to avoid vertices.  See also page 11-12
// (7-Sept-95) of MJD's micromagnetic NIST lab notebook.
int Polygon::IsInside(TwoPoint& pt)
{
  int intersections=0;
  for(int i=0;i<count-1;i++)
    intersections+=CrossSegment(pt,vertex[i],vertex[i+1]);
  intersections+=CrossSegment(pt,vertex[count-1],vertex[0]);
  return intersections%2;
}

// Polygon coordinate transformation routines (Scale, Shift and Rotate)
void Polygon::Scale(double xscale,double yscale)
{
  for(int i=0;i<count;i++) {
    vertex[i].x*=xscale;
    vertex[i].y*=yscale;
  }
}

void Polygon::Shift(double xoffset,double yoffset)
{
  for(int i=0;i<count;i++) {
    vertex[i].x+=xoffset;
    vertex[i].y+=yoffset;
  }
}

void Polygon::Rotate(double angcos,double angsin)
{
  double tempx,tempy;
  for(int i=0;i<count;i++) {
    tempx=vertex[i].x; tempy=vertex[i].y;
    vertex[i].x=angcos*tempx-angsin*tempy;
    vertex[i].y=angsin*tempx+angcos*tempy;
  }
}

// Rotates and rescales bitmap as necessary to shift vertex[0]
// to the nearest integral coordinate.  This may make some bitmaps
// prettier.
void Polygon::AdjustFirstToInteger(void)
{
  TwoPoint newpt;
  double magold,magnew,angcos,angsin,scale;
  magold=vertex[0].Mag();
  if(magold<EPSILON) return; // Pt[0] is too small to do anything with.
  newpt.x=ROUND(vertex[0].x);
  newpt.y=ROUND(vertex[0].y);
  magnew=newpt.Mag();
  scale=magnew/magold;
  angcos=(vertex[0].x*newpt.x+vertex[0].y*newpt.y)/(magold*magnew);
  angsin=sqrt(1-SQ(angcos));
  if(vertex[0].x*newpt.y<vertex[0].y*newpt.x) angsin*=-1;
  this->Rotate(angcos,angsin);
  this->Scale(scale,scale);
}

FileHeader::FileHeader(void)
{
  int i;
  for(i=0;i<(int)sizeof(Reserved);i++) Reserved[i]='\0';
  for(i=0;i<(int)sizeof(Note);i++)     Note[i]='\0';
}

int FileHeader::Verify(void)
{
  if(strcmp((const char *)Magic,VecFileMagic02)!=0)
    return VFH_BADMAGIC;
  if(EndianTest!=VecFileEndianTest &&
     flip(EndianTest)!=VecFileEndianTest)
    return VFH_UNKNOWNENDIAN;
  if(dataoffset!=sizeof(FileHeader))
    return VFH_BADOFFSET;
  if(strlen((const char *)Note)>=sizeof(Note))
    return VFH_BADNOTE;
  if(Nx<1 || Ny<1 || Nz<1 || dimen<1)
    return VFH_BADVALUES;
  return 0;
}

int FileHeader::Read(FILE* fptr,int& swapinbytes)
{
  if(fread((char *)Magic,sizeof(BYTE),sizeof(Magic),fptr)!=sizeof(Magic))
    return VF_PREMATUREEOF;
  if(fread((char *)&EndianTest,sizeof(REAL8),1,fptr)!=1)
    return VF_PREMATUREEOF;
  if(fread((char *)&Nx,sizeof(INT4),1,fptr)!=1)
    return VF_PREMATUREEOF;
  if(fread((char *)&Ny,sizeof(INT4),1,fptr)!=1)
    return VF_PREMATUREEOF;
  if(fread((char *)&Nz,sizeof(INT4),1,fptr)!=1)
    return VF_PREMATUREEOF;
  if(fread((char *)&dimen,sizeof(INT4),1,fptr)!=1)
    return VF_PREMATUREEOF;
  if(fread((char *)&dataoffset,sizeof(INT4),1,fptr)!=1)
    return VF_PREMATUREEOF;
  if(fread((char *)Reserved,sizeof(BYTE),sizeof(Reserved),fptr)
     !=sizeof(Reserved))
    return VF_PREMATUREEOF;
  if(fread((char *)Note,sizeof(BYTE),sizeof(Note),fptr)!=sizeof(Note))
    return VF_PREMATUREEOF;

  if(EndianTest==VecFileEndianTest) swapinbytes=0;
  else if(flip(EndianTest)==VecFileEndianTest) swapinbytes=1;
  else
    PlainError(1,"Error in FileHeader::Read: Unknown binary byte order");

  if(swapinbytes) {
    printf("Reversing binary byte order on input\n");
    Nx=flip(Nx); Ny=flip(Ny); Nz=flip(Nz);
    dimen=flip(dimen); dataoffset=flip(dataoffset);
  }
  return Verify();
}

int FileHeader::Write(FILE* fptr)
{
  if(fwrite((char *)Magic,sizeof(BYTE),sizeof(Magic),fptr)!=sizeof(Magic))
    return VF_PREMATUREEOF;
  if(fwrite((char *)&EndianTest,sizeof(REAL8),1,fptr)!=1)
    return VF_PREMATUREEOF;
  if(fwrite((char *)&Nx,sizeof(INT4),1,fptr)!=1)
    return VF_PREMATUREEOF;
  if(fwrite((char *)&Ny,sizeof(INT4),1,fptr)!=1)
    return VF_PREMATUREEOF;
  if(fwrite((char *)&Nz,sizeof(INT4),1,fptr)!=1)
    return VF_PREMATUREEOF;
  if(fwrite((char *)&dimen,sizeof(INT4),1,fptr)!=1)
    return VF_PREMATUREEOF;
  if(fwrite((char *)&dataoffset,sizeof(INT4),1,fptr)!=1)
    return VF_PREMATUREEOF;
  if(fwrite((char *)Reserved,sizeof(BYTE),sizeof(Reserved),fptr)
     !=sizeof(Reserved))
    return VF_PREMATUREEOF;
  if(fwrite((char *)Note,sizeof(BYTE),sizeof(Note),fptr)!=sizeof(Note))
    return VF_PREMATUREEOF;
  return Verify();
}

FileVector::FileVector(int psfontsize)
// psfontsize = PostScript font size in points
{
  fptr=NULL;
  Nx=Ny=Nz=dimen=cellsize=0;
  bitmap=(int **)NULL;
  TitleSize=psfontsize;
}

FileVector::~FileVector(void)
{
  if(fptr!=NULL) Close();
  if(bitmap!=(int **)NULL) {
    for(int i=0;i<bmheight;i++) delete bitmap[i];
    delete bitmap;
  }
}

void FileVector::SetNote(BYTE *newnote)
{
  int i;
  if(newnote!=NULL) {
    strncpy((char *)fh.Note,(char *)newnote,sizeof(fh.Note));
    fh.Note[sizeof(fh.Note)-1]='\0'; // Safety
  }
  else {
    for(i=0;i<(int)sizeof(fh.Note);i++) fh.Note[i]='\0';
  }
}

int FileVector::ReadHeader(const char *filename,
			   int &Xdimen,int &Ydimen,int &Zdimen,int &Vecdimen)
{
  if(filename == StandardInput) fptr=stdin;
  else if((fptr=Omf_FOpen(filename,"rb"))==NULL) {
    PlainWarning("Warning in FileVector::ReadHeader: %s",ErrBadFileOpen);
    return -1;
  }
  if(fh.Read(fptr,SwapInBytes)!=0) {
    PlainWarning("Warning in FileVector::ReadHeader: %s",ErrBadFile);
    return -2;
  }
  Xdimen=Nx=fh.Nx;
  Ydimen=Ny=fh.Ny;
  Zdimen=Nz=fh.Nz;
  Vecdimen=dimen=fh.dimen;
  dataoffset=fh.dataoffset;
  fseek(fptr,dataoffset,SEEK_SET);  // Reposition to start of data
  return 0;
}

int FileVector::WriteHeader(const char *filename,
			   int Xdimen,int Ydimen,int Zdimen,int Vecdimen)
{
  if(filename==StandardOutput) fptr=stdout;
  else if((fptr=Omf_FOpen(filename,"wb"))==NULL) {
    PlainWarning("Warning in FileVector::WriteHeader: %s",ErrBadFileOpen);
    return -1;
  }

  strncpy((char *)fh.Magic,VecFileMagic02,sizeof(fh.Magic)-1);
  fh.Magic[sizeof(fh.Magic)-1]='\0';  // Safety
  fh.EndianTest=VecFileEndianTest;
  fh.Nx=Nx=Xdimen;
  fh.Ny=Ny=Ydimen;
  fh.Nz=Nz=Zdimen;
  fh.dimen=dimen=Vecdimen;
  fh.dataoffset=dataoffset=sizeof(struct FileHeader);

  if(fh.Write(fptr)!=0) {
    PlainWarning("Warning in FileVector::WriteHeader: %s",ErrBadFile);
    return -2;
  }
  return 0;
}

int FileVector::ReadVecs(int count,double* arr)
{
#ifdef CODE_CHECKS
  if(sizeof(double)!=2*sizeof(int))
    PlainError("Error in FileVector::ReadVecs: "
	       "Code assumes data type double is twice as wide as int,"
	       " but sizeof(double)=%d, and sizeof(int)=%d",
	       sizeof(double),sizeof(int));
#endif
  if(fread((char *)arr,sizeof(double)*dimen,count,fptr)!=(unsigned)count)
    return VF_PREMATUREEOF;
  if(SwapInBytes) swap_words8(arr,count*dimen);

  return 0;
}

int FileVector::WriteVecs(int count,double* arr)
{
  if(fwrite((char *)arr,sizeof(double)*dimen,count,fptr)
     !=(unsigned)count)
    return VF_PREMATUREEOF;
  return 0;
}

int FileVector::WritePSHeader(const char *filename,
			      double Xdimen,double Ydimen,double Zdimen,
			      int Vecdimen,int drawtitle,char *subtitle)
{ 
  if(filename==StandardOutput) fptr=stdout;
  else if((fptr=Omf_FOpen(filename,"wb"))==NULL) {
    PlainWarning("Warning in FileVector::WritePSHeader: %s",ErrBadFileOpen);
    return -1;
  }
  Nx=(int)ceil(Xdimen);
  Ny=(int)ceil(Ydimen);
  Nz=(int)ceil(Zdimen);
  dimen=Vecdimen;

  // Calculate PS dimensions
  int landscape,titleheight;
  double xoffset,yoffset,pswidth,psheight,scale,scale2;
  double lmar,tmar,workwidth,workheight;

  if(!drawtitle) titleheight=0;
  else if(subtitle==NULL) titleheight=2*TitleSize;
  else                    titleheight=3*TitleSize;

  scale=MIN((PageWidth-2*PageLeftMargin)/Xdimen,  // Portrait scaling
	    (PageHeight-2*PageTopMargin-2*titleheight)/Zdimen);
  scale2=MIN((PageHeight-2*PageTopMargin)/Xdimen,  // Landscape scaling
	     (PageWidth-2*PageLeftMargin-2*titleheight)/Zdimen);
  if(scale>=scale2) {
    landscape=0;  // Portrait mode
    lmar=PageLeftMargin;
    tmar=PageTopMargin;
    workwidth=PageWidth;
    workheight=PageHeight;
  }
  else {
    landscape=1; // Landscape mode
    scale=scale2;
    lmar=PageTopMargin;
    tmar=PageLeftMargin;
    workwidth=PageHeight;
    workheight=PageWidth;
  }

  pswidth=Xdimen*scale;
  psheight=Zdimen*scale+titleheight;
  xoffset=lmar  // Skip past left margin
    +((workwidth-2*lmar)-pswidth)/2;  // Center in unfilled space
  yoffset=tmar  // Skip past bottom margin (==TopMargin)
    +((workheight-2*tmar)-psheight)/2;  // Center in unfilled space

  // PostScript file header
  fprintf(fptr,"%%!PS-Adobe-2.0\n");
  fprintf(fptr,"%%%%Title: %s\n",filename);
  fprintf(fptr,"%%%%Creator: FileVector C++ type, Michael J. Donahue 1995\n");
//  fprintf(fptr,"%%%%BoundingBox: 0 306 637 790\n");
  fprintf(fptr,"%%%%Pages: 1\n");
//  fprintf(fptr,"%%%%DocumentFonts:\n");
  fprintf(fptr,"%%%%EndComments\n");
  fprintf(fptr,"%%%%EndProlog\n\n");
  fprintf(fptr,"%%%%Page: 1 1\n\n");
  fprintf(fptr,"%% remember original state\n/origstate save def\n\n");
  fprintf(fptr,"%% build a temporary dictionary\n20 dict begin\n\n");

  // Switch to landscape mode if required
  if(landscape) fprintf(fptr,"%f 0 translate 90 rotate %% Landscape\n\n",
			PageWidth);

  // Draw title
  if(drawtitle) {
    fprintf(fptr,
	    "/Helvetica-Bold findfont %d scalefont setfont %% Title font\n",
	    TitleSize);
    fprintf(fptr,"(%s) dup stringwidth pop %% Center title\n",fh.Note);
    fprintf(fptr,"  %f exch sub 2 div %f moveto\n",
	    workwidth,yoffset+psheight-TitleSize);
    fprintf(fptr,"show %% Print title\n");
    if(subtitle!=NULL) { // Draw subtitle
      fprintf(fptr,"(%s) dup stringwidth pop %% Center subtitle\n",subtitle);
      fprintf(fptr,"  %f exch sub 2 div %f moveto\n",
	      workwidth,yoffset+psheight-2*TitleSize);
      fprintf(fptr,"show %% Print subtitle\n");
    }
  }

  // Initialize drawing parameters
  fprintf(fptr,"%f setlinewidth\n",ArrowLineWidth);
  fprintf(fptr,"%f %f translate\n",xoffset,yoffset);
  fprintf(fptr,"%f %f scale\n\n",scale,scale);

  // Draw outside box
  fprintf(fptr,"newpath %% Draw model enclosing box\n");
  fprintf(fptr,"0 0 moveto\n");
  fprintf(fptr,"%f 0 lineto\n",Xdimen);
  fprintf(fptr,"%f %f lineto\n",Xdimen,Zdimen);
  fprintf(fptr,"0 %f lineto\n",Zdimen);
  fprintf(fptr,"0 0 lineto\n");
  fprintf(fptr,"closepath\n");
  fprintf(fptr,"stroke\n\n");

  // Reset user coordinates to first cell
  fprintf(fptr,"0.5 0.5 translate %% Translate to first cell\n\n");

  // PostScript procedure to write an arrow from (-ArrowLength/2,0) 
  // to (ArrowLength/2,0), through the origin in direction dir.
  // Expects dir (in degrees) on top of stack
  fprintf(fptr,"/Arrow { %% stack: mag dir red green blue\n");
  fprintf(fptr,"  gsave\n");
  fprintf(fptr,"  setrgbcolor %% Arrow color\n");
  fprintf(fptr,"  rotate %% Arrow direction\n");
  fprintf(fptr,"  1 scale %% Arrow size (1=>don't rescale y, i.e. width)\n");
  fprintf(fptr,"  newpath\n");

  fprintf(fptr,"   %f 0 moveto\n",-ArrowLength/2.);
  fprintf(fptr,"   %f 0 lineto\n",(ArrowLength-ArrowHeadRatio)*0.5);
  fprintf(fptr,"  stroke\n"); 
  fprintf(fptr,"   %f 0 moveto\n",ArrowLength/2.);
  fprintf(fptr,"   %f %f lineto\n",ArrowLength/2.-ArrowHeadRatio*0.8,
	                           ArrowHeadRatio*0.3);
  fprintf(fptr,"   %f 0 lineto\n",(ArrowLength-ArrowHeadRatio)*0.5);
  fprintf(fptr,"   %f %f lineto\n",ArrowLength/2.-ArrowHeadRatio*0.8,
	                           -ArrowHeadRatio*0.3);
  fprintf(fptr,"   %f 0 lineto\n",ArrowLength/2.);

  fprintf(fptr,"  closepath\n");
  fprintf(fptr,"  fill\n");
  fprintf(fptr,"  grestore\n");
  fprintf(fptr,"} def\n\n");

  // PostScript procedure to draw a circle and dot at the origin
  fprintf(fptr,"/DotCircle { %% stack: red green blue\n");
  fprintf(fptr,"  setrgbcolor %% Color\n");
  fprintf(fptr,"  gsave\n");
  fprintf(fptr,"  newpath\n");
  fprintf(fptr,"  %f setlinewidth\n",ArrowLineWidth/2.0);
  fprintf(fptr,"  0 0 0.20 0 360 arc stroke\n");
  fprintf(fptr,"  0 0 0.05 0 360 arc fill\n");
  fprintf(fptr,"  grestore\n");
  fprintf(fptr,"} def\n\n");

  // PostScript procedure to draw a circle and cross (x) at the origin
  fprintf(fptr,"/CrossCircle { %% stack: red green blue\n");
  fprintf(fptr,"  setrgbcolor %% Color\n");
  fprintf(fptr,"  gsave\n");
  fprintf(fptr,"  newpath\n");
  fprintf(fptr,"  %f setlinewidth\n",ArrowLineWidth/2.0);
  fprintf(fptr,"  0 0 0.2 0 360 arc\n");
  fprintf(fptr,"  -0.1414 -0.1414 moveto 0.1414  0.1414 lineto\n");
  fprintf(fptr,"  -0.1414  0.1414 moveto 0.1414 -0.1414 lineto\n");
  fprintf(fptr,"  closepath stroke\n");
  fprintf(fptr,"  grestore\n");
  fprintf(fptr,"} def\n\n");

  // Postscript position variables
  fprintf(fptr,"/row 0 def %% Current row\n");
  fprintf(fptr,"/rowstop %d def %% Number of rows\n\n",Nz);

  // Move to next cell
  fprintf(fptr,"\n/SkipCell { %% Increment row and Move to next cell\n");
  fprintf(fptr,  "  /row row 1 add def  %% Increment row\n");
  fprintf(fptr,  "  row rowstop lt      %% row < rowstop?\n");
  fprintf(fptr,  "    { 0 1 translate } %% Shift up\n");
  fprintf(fptr,  "    { 1 1 rowstop sub translate %% Right and down\n");
  fprintf(fptr,  "      /row 0 def } ifelse %% Reset row\n");
  fprintf(fptr,  "} def\n\n");

  // Do one arrow cell
  fprintf(fptr,"\n/DoArrowCell { %% Does one arrow cell and increments row\n");
  fprintf(fptr,  "  Arrow   %% Stack: mag dir red green blue\n");
  fprintf(fptr,  "  SkipCell %% Move to next cell\n");
  fprintf(fptr,  "} def\n\n");

  // Do one dot-circle cell
  fprintf(fptr,"\n/DoDotCircleCell { %% Do one dot-circle cell & inc. row\n");
  fprintf(fptr,  "  DotCircle   %% Stack: red green blue\n");
  fprintf(fptr,  "  SkipCell %% Move to next cell\n");
  fprintf(fptr,  "} def\n\n");

  // Do one cross-circle cell
  fprintf(fptr,"\n/DoCrossCircleCell { %% Do 1 x-circle cell & inc. row\n");
  fprintf(fptr,  "  CrossCircle   %% Stack: red green blue\n");
  fprintf(fptr,  "  SkipCell %% Move to next cell\n");
  fprintf(fptr,  "} def\n\n");

  return 0;
}

int FileVector::WritePSVecs(int count,double* arr,
			    int comp0,int comp1,int compcolor)
{ // comp0=>x component, comp1=>y component
  double dir,mag,color,red,blue,x,y;
  while(count-- >0) {
    x=arr[comp0]; y=arr[comp1];
    if(compcolor>=0) color=arr[compcolor]; else color=0.0;
    arr+=dimen;
    mag=SQ(x)+SQ(y);
    if(mag>EPSILON) {
      dir=atan2(y,x);
      dir*=180./PI; // Convert to degrees
    }
    else dir=0.0;
    if(color>=0) { blue=0.; red=MIN(1.,color); }
    else         { red=0.;  blue=MIN(1.,-color); }
    if(mag>VertMagCrossover)
      fprintf(fptr,"%f %8.2f %f 0.0 %f DoArrowCell\n",mag,dir,red,blue);
    else if(color>0)
      fprintf(fptr,"%f 0.0 %f DoCrossCircleCell\n",red,blue);
    else
      fprintf(fptr,"%f 0.0 %f DoDotCircleCell\n",red,blue);
  }
  return 0;
}

int FileVector::WritePSTrailer(void)
{
  fprintf(fptr,"\nshowpage\n");
  fprintf(fptr,"%% stop using temporary dictionary\nend\n\n");
  fprintf(fptr,"%% restore original state\norigstate restore\n\n");
  fprintf(fptr,"%%%%Trailer\n");
//  fprintf(fptr,"%%Trailer\n%%EOF\n%c\n",0x04);
    // NOTE: tgif wants a newline trailing the Ctrl-D character
  return 0;
}

int FileVector::WritePPMHeader(const char *filename,
			       int Xdimen,int Ydimen,int Zdimen,
			       int Vecdimen,int Imagecellsize)
{ // The important dimenstions are Xdimen and Zdimen (width & height, resp.)
  Nx=Xdimen;  Ny=Ydimen;  Nz=Zdimen;  dimen=Vecdimen;
  cellsize=Imagecellsize;

  // Allocate bitmap for 1 output row
  bmwidth=cellsize*Nx;    bmheight=cellsize;
  if(!(bitmap=new int*[bmheight]))
    PlainError(1,"Error in FileVector::WritePPMHeader: %s",ErrNoMem);
  for(int i=0;i<bmheight;i++) {
    if(!(bitmap[i]=new int[bmwidth]))
      PlainError(1,"Error in FileVector::WritePPMHeader: %s",ErrNoMem);
  }

  // Initialize base bitmap arrow
  // It is similar to the arrow defined above for PostScript output
  basearrow.Resize(ArrowVertexCount);
  workarrow.Resize(ArrowVertexCount);
  // First set up for a 1x1 cell (ASSUMES ArrowVertexCount=8)
  basearrow[0].x= -ArrowLength/2.;
  basearrow[0].y=BitmapLineWidth/2.;
  basearrow[1].x=ArrowLength/2.-BitmapArrowHeadRatio*0.8;
  basearrow[1].y=basearrow[0].y;
  basearrow[2].x=ArrowLength/2.-BitmapArrowHeadRatio;
  basearrow[2].y=BitmapArrowHeadRatio*0.4;
  basearrow[3].x=ArrowLength/2.;
  basearrow[3].y=BitmapLineWidth*0.5;
  basearrow[4].x=  basearrow[3].x;
  basearrow[4].y= -basearrow[3].y;
  basearrow[5].x=  basearrow[2].x;
  basearrow[5].y= -basearrow[2].y;
  basearrow[6].x=  basearrow[1].x;
  basearrow[6].y= -basearrow[1].y;
  basearrow[7].x=  basearrow[0].x;
  basearrow[7].y= -basearrow[0].y;
  // Rescale to actual cell size
  basearrow.Scale(cellsize,cellsize);

  // Initialize base bitmap cross
  basecross.Resize(CrossVertexCount);
  // First set up a '+' inside a 1x1 cell (ASSUMES CrossVertexCount=12)
  basecross[0].x=BitmapCircleRadius-BitmapLineWidth/2.; // Approximate
  basecross[0].y=BitmapLineWidth/2.;
  basecross[1].x=basecross[1].y=basecross[0].y;
  basecross[2].x=   basecross[0].y;
  basecross[2].y=   basecross[0].x;
  basecross[3].x=  -basecross[2].x;
  basecross[3].y=   basecross[2].y;
  basecross[4].x=  -basecross[1].x;
  basecross[4].y=   basecross[1].y;
  basecross[5].x=  -basecross[0].x;
  basecross[5].y=   basecross[0].y;
  basecross[6].x=  -basecross[0].x;
  basecross[6].y=  -basecross[0].y;
  basecross[7].x=  -basecross[1].x;
  basecross[7].y=  -basecross[1].y;
  basecross[8].x=  -basecross[2].x;
  basecross[8].y=  -basecross[2].y;
  basecross[9].x=   basecross[2].x;
  basecross[9].y=  -basecross[2].y;
  basecross[10].x=  basecross[1].x;
  basecross[10].y= -basecross[1].y;
  basecross[11].x=  basecross[0].x;
  basecross[11].y= -basecross[0].y;
  // Rotate 45 degrees
  basecross.Rotate(0.70710678,0.70710678);
  // Rescale to actual cell size
  basecross.Scale(cellsize,cellsize);

  // Open file and write header
  if(filename==StandardOutput) fptr=stdout;
  else if((fptr=Omf_FOpen(filename,"wb"))==NULL) {
    PlainWarning("Warning in FileVector::WritePPMHeader: %s",ErrBadFileOpen);
    return -1;
  }
  setvbuf(fptr,(char *)NULL,_IOFBF,8192);
  fprintf(fptr,"P3\n%d %d\n%d\n",Nx*cellsize,Nz*cellsize,MaxColorValue);

  return 0;
}

void FileVector::DrawBitmapLine(int r1,int c1,int r2,int c2,int color)
{ // Low level line drawing routine
  int rdiff,cdiff,count,i,r,c;
  rdiff=r2-r1; cdiff=c2-c1; count=MAX(abs(rdiff),abs(cdiff));
  if(count==0) bitmap[r1][c1]=color;
  else {
    for(i=0;i<=count;i++) {
      r=r1+(rdiff*i+count/2)/count;
      c=c1+(cdiff*i+count/2)/count;
      bitmap[r][c]=color;
    }
  }
}

void FileVector::DrawBitmapArrow(int xcoor,
				 double len,double angcos,double angsin,
				 int color)
{
  int r,c,coffset;
  TwoPoint pt;
  // Setup appropriately sized and rotated bitmap arrow
  if(len<EPSILON) return; // This should be a dot or cross
  workarrow=basearrow; // Make working copy of base arrow
  workarrow.Scale(len,1.);  // Squash along long axis
  workarrow.Rotate(angcos,angsin);
  workarrow.AdjustFirstToInteger(); // Shift to integer grid
  coffset=xcoor*cellsize;
  // Step through each point in cell,
  // and paint if it is inside the arrow
  for(r=0;r<cellsize;r++) for(c=0;c<cellsize;c++) {
    pt.x=c-(cellsize-1)/2.;
    pt.y=(cellsize-1)/2.-r;
    if(workarrow.IsInside(pt)) bitmap[r][c+coffset]=color;
  }
}

void FileVector::DrawBitmapCircleCross(int xcoor,int color)
{
  int r,c,coffset;
  double cdistsq,radsqouter,radsqinner;
  TwoPoint pt;
  coffset=xcoor*cellsize;
  radsqouter=BitmapCircleRadius*cellsize;
  radsqouter*=radsqouter;
  radsqinner=(BitmapCircleRadius-BitmapLineWidth)*cellsize;
  radsqinner*=radsqinner;
  // Step through each point in cell,
  // and paint if needed
  for(r=0;r<cellsize;r++) for(c=0;c<cellsize;c++) {
    pt.x=c-(cellsize-1)/2.;
    pt.y=(cellsize-1)/2.-r;
    cdistsq=SQ(pt.x)+SQ(pt.y);
    if(cdistsq>radsqouter) continue; // Outside outer circle
    if(cdistsq>radsqinner || basecross.IsInside(pt))
      bitmap[r][c+coffset]=color;
  }
}

void FileVector::DrawBitmapCircleDot(int xcoor,int color)
{
  int r,c,coffset;
  double cdistsq,radsqouter,radsqinner,radsqdot;
  TwoPoint pt;
  coffset=xcoor*cellsize;
  radsqouter=BitmapCircleRadius*cellsize;
  radsqouter*=radsqouter;
  radsqinner=(BitmapCircleRadius-BitmapLineWidth)*cellsize;
  radsqinner*=radsqinner;
  radsqdot=BitmapDotRadius*cellsize;
  radsqdot*=radsqdot;
  // Step through each point in cell,
  // and paint if needed
  for(r=0;r<cellsize;r++) for(c=0;c<cellsize;c++) {
    pt.x=c-(cellsize-1)/2.;
    pt.y=(cellsize-1)/2.-r;
    cdistsq=SQ(pt.x)+SQ(pt.y);
    if(cdistsq>radsqouter) continue; // Outside outer circle
    if(cdistsq>radsqinner || cdistsq<radsqdot)
      bitmap[r][c+coffset]=color;
  }
}


void FileVector::DrawBitmapDot(int xcoor,int color)
{
  int r,c,coffset;
  double cdistsq,radsqdot;
  TwoPoint pt;
  coffset=xcoor*cellsize;
  radsqdot=BitmapDotRadius*cellsize;
  radsqdot*=radsqdot;
  // Step through each point in cell,
  // and paint if needed
  for(r=0;r<cellsize;r++) for(c=0;c<cellsize;c++) {
    pt.x=c-(cellsize-1)/2.;
    pt.y=(cellsize-1)/2.-r;
    cdistsq=SQ(pt.x)+SQ(pt.y);
    if(cdistsq<radsqdot) bitmap[r][c+coffset]=color;
  }
}


int FileVector::WritePPMVecs(int count,double* arr,
			     int comp0,int comp1,int compcolor,
			     int colormap)
{ // Color sectants: 0=Red, 1=Yellow, 2=Green, 3=Cyan, 4=Blue, 5=Magenta
  // NOTE: Count must be 1 row of vectors == Nx
  double xcomp,zcomp,mag;
  int r,c,color;
  if(count!=Nx)
    PlainError(1,"Error in FileVector::WritePPMVecs: %s",ErrBadParam);

  // Erase bitmap
  for(r=0;r<bmheight;r++) for(c=0;c<bmwidth;c++)
    bitmap[r][c]=MaxColorValue+1; // Erase

  // Fill bitmap
  for(int i=0;i<count;i++,arr+=dimen) {
    xcomp=arr[comp0]; zcomp=arr[comp1];
    mag=hypot(xcomp,zcomp);
    if(compcolor>=0) {
      color=(int)ROUND(MaxColorValue*arr[compcolor]);
      if(color>MaxColorValue)       color=MaxColorValue;
      else if(color<-MaxColorValue) color=-MaxColorValue;
    }
    else color=0;
    if(mag>VertMagCrossover && mag*cellsize>BitmapArrowMinPixels)
      DrawBitmapArrow(i,mag,xcomp/mag,zcomp/mag,color);
//    else if(color>0) DrawBitmapCircleCross(i,color);
//    else             DrawBitmapCircleDot(i,color);
    else DrawBitmapDot(i,color);
  }


  // Dump bitmap
  int red,green,blue;  double R,G,B,hue;
  if(colormap==2) {     // blue -> black -> red, on white background
    for(r=0;r<bmheight;r++) for(c=0;c<bmwidth;c++) {
      color=bitmap[r][c];
      if(color>MaxColorValue) red=green=blue=MaxColorValue; // White background
      else {
	green=0;
	if(color>0) { red=color;   blue=0; }
	else        { blue=-color; red=0;  }
      }
      fprintf(fptr,"%3d %3d %3d\n",red,green,blue);
    }
  }
  else if(colormap==1) { // green -> blue -> red, on white background
    for(r=0;r<bmheight;r++) for(c=0;c<bmwidth;c++) {
      color=bitmap[r][c];
      if(color>MaxColorValue) { // White background
	red=blue=green=MaxColorValue;
      }
      else {
	hue=(4./6.) // Blue
	  + (double(color)/double(MaxColorValue))/3.; // Green->Blue->Red
	HSVtoRGB(hue,1.0,0.9,R,G,B);   // Slightly darkened
	red=(int)ROUND(MaxColorValue*R); 
	green=(int)ROUND(MaxColorValue*G);
	blue=(int)ROUND(MaxColorValue*B);
      }
      fprintf(fptr,"%3d %3d %3d\n",red,green,blue);
    }
  }
  else { // Default colormap: yellow -> green -> cyan, on magenta background
    for(r=0;r<bmheight;r++) for(c=0;c<bmwidth;c++) {
      color=bitmap[r][c];
      if(color>MaxColorValue) { // Magenta background
	green=0;
	red=blue=(int)ROUND(MaxColorValue*BACKBRIGHT);
      }
      else {
	hue=(1./3.) // Green
	  + (double(color)/double(MaxColorValue))/6.; // Yellow->Green->Cyan
	HSVtoRGB(hue,1.0,1.0,R,G,B);
	red=(int)ROUND(MaxColorValue*R); 
	green=(int)ROUND(MaxColorValue*G);
	blue=(int)ROUND(MaxColorValue*B);
      }
      fprintf(fptr,"%3d %3d %3d\n",red,green,blue);
    }
  }

  return 0;
}

int FileVector::WriteColorPPMHeader(const char *filename,
				    int Xdimen,int Ydimen,int Zdimen,
				    int Vecdimen,double max)
{
  maxval=max;
  Nx=Xdimen;  Ny=Ydimen;  Nz=Zdimen;  dimen=Vecdimen;

  // Open file and write header
  fflush(stdout);
  if(filename==StandardOutput) fptr=stdout;
  else {
    if((fptr=Omf_FOpen(filename,"wb"))==NULL)
    PlainError(1,"Error in FileVector::WriteColorPPMHeader:"
	       " Unable to open %s for output",filename);
  }

  // Write header
  fprintf(fptr,"P3\n%d %d\n%d\n",Nx,Nz,MaxGrey);
  return 0;
}

int FileVector::WriteColorPPMPix(int count,double* arr)
{
  int i,j,val;
  double offset=0.0;
  double scale=double(MaxGrey)/double(maxval);
  for(i=0;i<count;i++) {
    for(j=0;j<dimen;j++,arr++) {
      val=(int)ROUND(((*arr)+offset)*scale);
      fprintf(fptr,"%d ",val);
    }
    fprintf(fptr,"\n"); // No more than 70 characters per line
                        // are allowed by PPM standard.
  }
  return 0;
}

void FileVector::WriteColorPPMClose(void)
{ // Call this _INSTEAD_ of Close()
  Close();
}

void FileVector::Close(void)
{
  if(fptr!=(FILE *)NULL) {
    fclose(fptr);  fptr=NULL;
  }
  if(bitmap!=(int **)NULL) {
    for(int i=0;i<bmheight;i++) delete bitmap[i];
    delete bitmap;
    bitmap=(int **)NULL;
  }
}

