# FILE: mmdsavedlg.tcl
#
# Save file dialog box for mmDisp
#
# Callbacks: SAVE
#            DELETE $delete_callback_arg
#
# debugging enable/disable/level

em_class omf_mmdsavedlg {
    const public variable winpath
    const public variable outer_frame_options = "-bd 2 -relief raised"
    const public variable filter = "*.o?f *.svf *.vio"
    const public variable dialog_title = "Save File"
    const public variable mesh_update_proc = em_nop
    const public variable mesh_update_variable = {} ;# If non-nil, then
    ## a w trace gets setup on this global (name) to automatically
    ## update "Regular" gridtype availability, and also file title
    ## and description fields.  New values are obtained by calling
    ## [eval $mesh_update_proc cget -type], ...-title, and 
    ## ...-description.  The returned gridtype should be one of
    ## "GridVec3f", "GeneralVec3f", or "Empty".
    public variable callback = em_nop
    public variable delete_callback_arg = {}
    private variable filewidget = {}
    private variable optsbox = {}
    private variable ctrlwidget = {}
    private variable descbox = {}
    const private variable titlelabel = "Title:"
    const private variable desclabel = "Desc:"
    private variable selectionwidget = {}
    private variable dirwidget = {}
    constructor {args} {
	eval $this configure -winpath .w$this $args
	# Create toplevel frame to hold dialog
	toplevel $winpath
	wm group $winpath .
	wm title $winpath $dialog_title
	if {![string match {} [info commands OMF_PositionChild]]} {
	    OMF_PositionChild $winpath  ;# Position at +.25+.25 over '.'
	}
	wm protocol $winpath WM_DELETE_WINDOW \
		"$this SwitchBoard $this CANCEL"

	# Temporarily disable callbacks until construction is complete
	set callback_keep $callback ;	set callback em_nop

	# Create subwidgets
	set filewidget [omf_fileselect new $winpath \
		-callback "$this SwitchBoard" \
		-filter $filter]
	set ctrlframe [frame $winpath.ctrlframe]
	set optsbox [omf_mmdomfopts new $winpath]
	set ctrlwidget [omf_ctrlwbrowse new $winpath \
		-callback "$this SwitchBoard"]
	set descbox [omf_multibox new $winpath -labelwidth 5 \
		-valuewidth 10 -labellist "$titlelabel {} $desclabel {}"]
	set selectionwidget [omf_selectbox new $winpath \
		-title "Save File As..." \
		-callback "$this SwitchBoard"]
	set dirwidget [omf_dirselect new $winpath \
		-callback "$this SwitchBoard"]

	# Initialize selection path
	$selectionwidget SetPath [$dirwidget cget -value]

	# Pack subwidgets
	pack [$filewidget cget -winpath] -side left \
		-fill both -expand 1 -padx 2m -pady 2m
	pack [$ctrlwidget cget -winpath] -in $ctrlframe \
		-side right -expand 0 -anchor ne -padx 2m -pady 2m
	pack [$optsbox cget -winpath] -in $ctrlframe \
		-side top -expand 0 -padx 2m -pady 2m
	pack $ctrlframe -side top \
		-fill x -expand 0 -anchor ne -padx 2m -pady 2m
	pack [$descbox cget -winpath] -side top \
		-fill x -expand 0 -anchor ne -padx 2m -pady 2m
	pack [$selectionwidget cget -winpath] -side top \
		-fill x -expand 0 -anchor n -padx 2m -pady 2m
	pack [$dirwidget cget -winpath] -side right \
		-fill both -expand 1 -padx 2m -pady 2m

	# Widget bindings
	bind $winpath <Destroy> "+$this WinDestroy %W"

	# Mesh (grid) types setup and trace
	if { ![ string match {} mesh_update_variable ] } {
	    upvar #0 $mesh_update_variable meshchange
	    if {![info exists meshchange]} { set meshchange {} }
	    $this MeshChangeTrace $mesh_update_variable {} w
	    trace variable meshchange w "$this MeshChangeTrace"
	}

	# Re-enable callbacks
	set callback $callback_keep
    }
    method MeshChangeTrace { name elt op } { \
	    mesh_update_variable mesh_update_proc \
	    optsbox descbox titlelabel desclabel } {
	set gridtype [eval $mesh_update_proc cget -gridtype]
	if { [string match "GridVec3f" $gridtype] } {
	    $optsbox GridsAllowedUpdate 1
	} else {
	    $optsbox GridsAllowedUpdate 0
	}
	$descbox Set $titlelabel [eval $mesh_update_proc cget -title]
	$descbox Set $desclabel [eval $mesh_update_proc cget -description]
    }
    method GetOptionState {} { optsbox } {
	# Returns a list of option box state pairs,
	# e.g., -datatype binary4 -gridtype regular
	return [$optsbox GetState]
    }
    method GetDescStrings {} { descbox titlelabel desclabel } {
	# Returns a list of description string pairs,
	# e.g., -title "This is the title" -desc "Any description"
	set desclist {}
	lappend desclist -title [$descbox LookUp $titlelabel]
	lappend desclist -desc [$descbox LookUp $desclabel]
	return $desclist
    }
    method GetFilename {} { selectionwidget } {
	# Returns full pathname of selected value
	return [$selectionwidget ReadFilename]
    }
    method SwitchBoard { subwidget actionid args } \
	    { filewidget selectionwidget dirwidget callback ctrlwidget } {
	# General callback routing routine
	switch -exact -- $actionid {
	    UPDATE {
		# Update actions
		if {[string compare $subwidget $filewidget]==0} {
			$selectionwidget SetFullname [lindex $args 1]
		} elseif {[string compare $subwidget $dirwidget]==0} {
			$filewidget SetPath [lindex $args 1]
			$selectionwidget SetPath [lindex $args 1]
		} else {
		    error "Unknown UPDATE callback source: [\
			    $subwidget class]"
		}
	    }
	    SELECT {
		set errmsg [eval $callback $this SAVE]
		if { ![$ctrlwidget cget -browse_state] \
			&& [string match {} $errmsg] } {
		    $this delete ; return
		}
		$filewidget Rescan
		$dirwidget Rescan
	    }
	    RESCAN {
		$filewidget Rescan
		$dirwidget Rescan
	    }
	    CANCEL {
		$this delete ; return
	    }
	    DELETE {}
	    default { error "Unknown actionid: $actionid" }
	}
    }
    method WinDestroy { w } { winpath } {
	if {[info exists winpath] && [string compare $winpath $w]!=0} {
	    return    ;# Child destroy event
	}
	$this delete
    }
    private variable delete_in_progress = 0
    destructor {
	# Protect against re-entrancy
	if { [info exists delete_in_progress] && \
		$delete_in_progress } { return }
	set delete_in_progress 1

	# Notify client of impending destruction
	catch {eval $callback $this DELETE {$delete_callback_arg}}

	# Remove meshtype trace
	catch {
	    if { ![ string match {} mesh_update_variable ] } {
		upvar #0 $mesh_update_variable meshchange
		trace vdelete meshchange w "$this MeshChangeTrace"
	    }
	}

	# Delete subwidgets
	catch {$filewidget      delete}
	catch {$ctrlwidget      delete}
	catch {$selectionwidget delete}
	catch {$dirwidget       delete}

	# Delete window
	if {[info exists winpath] && [winfo exists $winpath]} {
	    # Destroy instance frame, all children, and bindings
	    destroy $winpath
	}
    }
}
