/* FILE: omfsh.cc                 -*-Mode: c++-*-
 *
 *	The Tcl shell for the OOMMF project.
 *
 */

#include <tcl.h>
#include <tk.h> 
#include <string.h>
#include "machine.h"
#include "misc.h"	/* Only for panic() prototype */
#include "oommf.h"
/* End includes */     

/*
 * Do we load Tk? (Default = yes)
 */
static BOOL useTk = 1;

/*
 * Do we fork a child process to do the work? (Default = yes)
 *  -- Meaningful only when (SYSTEM_TYPE == UNIX)
 */
static BOOL forkflag = 1;

/* 
 * Removes argument "remove_index" from argument list by
 * translating all arguments after it down one spot and
 * decrementing argc.
 */
static void 
remove_arg(int &argc, char **argv, int remove_index)
{ 
    if(remove_index<0 || remove_index>=argc) return;
    for(int j=remove_index+1;j<argc;j++) {
        argv[j-1]=argv[j];
    }
    argc--;
}

/*
 * Sets static BOOL variables useTk and forkflag from command line
 */
static void 
ExtractOptions(int &argc, char **argv)
{
    int i=1;
    while(i<argc) {
        if(strcmp(argv[i],"--")==0){
            // End flag processing
            break;
        }

        if(argv[i][0]!='-') {
            // Not an option flag. Under Windows, should one
            // use '/' instead of '-' ??? 
            i++;
            continue;
        }

        char *optflag=&(argv[i][1]);
        if(*optflag=='-') optflag++;  // Treat "-" and "--" the same

        // Known options
        //	Should this processing be case sensitive ???
        if(strcmp(optflag,"tk")==0) {
            useTk=1;
            remove_arg(argc,argv,i);
        }
        else if(strcmp(optflag,"notk")==0 || strcmp(optflag,"no_tk")==0) {
            useTk=0;
            remove_arg(argc,argv,i);
        }
        else if(strcmp(optflag,"fork")==0) {
            forkflag=1;
            remove_arg(argc,argv,i);
        }
        else if(strcmp(optflag,"nofork")==0 || strcmp(optflag,"no_fork")==0) {
            forkflag=0;
            remove_arg(argc,argv,i);
        }
        else {
            // Anything else is left on the command line untouched, for
            // possible use by the shell script
            i++;
        }
    }
}

/* 
 * In Tk 4.2 (which is paired with Tcl 7.6), there is a bug in the 
 * function Tk_Init.  This short script patches over the problem.
 * It has to be placed here so that it can be evaluated between the
 * calls to Tcl_Init and Tk_Init.
 */
static char tk42InitPatch[] =
"if {[string match 7.6 $tcl_version] && ![info exists tk_library]} {\n\
   set tk_library [file join [file dirname $tcl_library] tk4.2]\n\
 }";

/* 
 * The "application initialization" for OOMMF applications.
 */
int 
Tcl_AppInit(Tcl_Interp *interp)
{
    // NOTE: The calls to panic() in this routine are admittedly a rather
    // crude means of error reporting, but it is the only means which is
    // reliably available across all the platforms and Tcl versions we
    // support at this early point in the execution of the application.  
    // Only after successful completion of the call to Oommf_Init below
    // can we be sure that all the necessary cross-platform building blocks
    // are in place for more sophisticated error reporting.
    //
    // If at some point in the future we abandon support for pre-8.0
    // versions of Tcl/Tk, then calls to panic may no longer be the
    // least common denominator.

  AutoBuf ab;  // Buffer space for calls to Tcl_Eval and panic.
#if (SYSTEM_TYPE==WINDOWS)
    // Because the entry point of this program is the function
    // WinMain() rather than main(), Windows will not provide the
    // standard channels stdin, stdout, and stderr.  Furthermore,
    // the Tcl library will not be able to set up the 
    // corresponding default channels.  The downside of this is that
    // attempts to operate on the standard channels in Tcl code will
    // cause errors.  Here we avoid that problem by opening the
    // "file" NUL: --  the null device on Windows platforms -- and
    // creating channels from it to be used as the standard channels
    // A further improvement would be to find a way to make the
    // standard Tcl channels and the C++ variables stdin, stdout, and
    // stderr refer to the same devices.  For now, avoiding fatal
    // errors is sufficient.
    // NOTE: This code chunk has to be here rather than in WinMain
    // because the Tcl library function calls require the interpreter
    // as an argument, and the interpreter is created in the Tcl_Main
    // or Tk_Main routine.
    Tcl_Channel channel;
    channel = Tcl_OpenFileChannel(interp, "NUL:", "r", 0666);
    if (channel == NULL) {
        panic("Couldn't create a standard input channel");
    }
    Tcl_SetStdChannel(channel, TCL_STDIN);
    channel = Tcl_OpenFileChannel(interp, "NUL:", "a", 0666);
    if (channel == NULL) {
        panic("Couldn't create a standard output channel");
    }
    Tcl_SetStdChannel(channel, TCL_STDOUT);
    channel = Tcl_OpenFileChannel(interp, "NUL:", "a", 0666);
    if (channel == NULL) {
        panic("Couldn't create a standard error channel");
    }
    Tcl_SetStdChannel(channel, TCL_STDERR);
#endif
 
    if (Tcl_Init(interp) != TCL_OK) {
      panic(ab("Tcl initialization error:\n\n%s\n\n\
As a workaround, set the environment variable\n\
TCL_LIBRARY to the name of a directory\n\
containing an error-free init.tcl."),interp->result);
    }
    if (useTk) {
        if (Tcl_Eval(interp, tk42InitPatch) != TCL_OK) {
	  panic(ab("Tk initialization error:\n\n%s \
(patch script for Tk 4.2 initialization bug)."),interp->result);
        }
        if (Tk_Init(interp) != TCL_OK) {
	  panic(ab("Tk initialization error:\n\n%s\n\n\
As a workaround, set the environment variable\n\
TK_LIBRARY to the name of a directory\n\
containing an error-free tk.tcl."),interp->result);
        }

        // Set up so that [info loaded] result includes Tk
        // Must be done after call to Tk_Init, but before call to Oommf_Init
        Tcl_StaticPackage(interp,ab("Tk"), Tk_Init,
			  (Tcl_PackageInitProc *) NULL);
    }
    if (Oommf_Init(interp) != TCL_OK) {
      panic(ab("Oommf initialization error:\n\n%s"),interp->result);
    }
    return TCL_OK;
}

#if (SYSTEM_TYPE==WINDOWS)
/*
 * C functions specific to the Windows platforms.
 */

/*
 * Routine to display panic() error message before dying 
 */
static void
WinPanic TCL_VARARGS_DEF(char *,arg1)
{
    va_list argList;
    char buf[4096];
    char *format;

    format = TCL_VARARGS_START(char *,arg1,argList);
    vsprintf(buf, format, argList);

    MessageBeep(MB_ICONEXCLAMATION);
    MessageBox(NULL, buf, "OOMMF Error",
            MB_ICONSTOP | MB_OK | MB_TASKMODAL | MB_SETFOREGROUND);
    ExitProcess(1);
}

/*
 * The main entry point for OOMMF applications on Windows platforms.
 */
int APIENTRY 
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine,
        int nCmdShow)
{
    // Since there's no stderr, give the error messages someplace to go
    Tcl_SetPanicProc(&WinPanic);

    // Windows doesn't pass the application name as the first word on the
    // command line like Unix does.  Instead a system call provides the
    // application name (as a full pathname).
    // NOTE: This fixed-length buffer should be long enough even for a full
    // pathname.  Dynamic allocation from the heap is avoided since
    // Tcl_Main and Tk_Main never return, and we'll have no chance to 
    // free the memory.  It's not clear whether Windows will free memory
    // when the process exits.
    char *p,exename[1024];
    GetModuleFileName(NULL, exename, sizeof(exename));
    // Translate '\' to '/'
    for (p = exename; *p != '\0'; p++) {
	if (*p == '\\') *p = '/';
    }
    for (p = lpszCmdLine; *p != '\0'; p++) {
	if (*p == '\\') *p = '/';
    }

    // Windows doesn't tokenize the command line
    int argc;
    char *argv[256],**argv_temp;
    char cmdline[4096];
    if(Tcl_SplitList(NULL,lpszCmdLine,&argc,&argv_temp)!=TCL_OK) {
      WinPanic("Bad command line: %s",lpszCmdLine);
    }
    if(argc+1>sizeof(argv)/sizeof(argv[0])) {
      WinPanic("Too many command line parameters: %d>%d",
	       argc+1,sizeof(argv)/sizeof(argv[0]));
    }

    // Copy split string out of temporary memory, and setup
    // argv pointers.
    argv[0]=exename;
    int i,total_length=0;
    for(i=0;i<argc;i++) {
      argv[i+1]=cmdline+total_length;
      total_length+=strlen(argv_temp[i])+1;
      if(total_length>sizeof(cmdline)) {
	WinPanic("Parsed command line too long (%d)",sizeof(cmdline));
      }
      strcpy(argv[i+1],argv_temp[i]);
    }
    argc++;
    omf_Tcl_Free((char *)argv_temp);

    // Now we can pretend a Unix system gave us argc and argv
    ExtractOptions(argc,argv);

    // Turn over control to the Tcl/Tk library.
    if (useTk) {
        Tk_Main(argc, argv, Tcl_AppInit);
    } else {
        Tcl_Main(argc, argv, Tcl_AppInit);
    }

    // Make picky compilers happy
    return 0;
}

#elif (SYSTEM_TYPE == UNIX)
/*
 * C functions specific to Unix platforms.
 */

/*
 * The main entry point for OOMMF applications on Unix platforms.
 */
int 
main(int argc, char **argv)
{
    ExtractOptions(argc,argv);

    // Fork off a child process to do all the work.
    // Parent process terminates and user gets back shell prompt
    if(!forkflag || fork()==0) { // Fork only occurs if forkflag!=0
        // Child or non-forked process

        // In child process, disable stdin, stdout
        if (forkflag) {
            freopen("/dev/null","r",stdin);
            freopen("/dev/null","a",stdout);
        }

        // Turn over control to the Tcl/Tk library.
        if (useTk) {
            Tk_Main(argc, argv, Tcl_AppInit);
        } else {
            Tcl_Main(argc, argv, Tcl_AppInit);
        }
    }

    // Make picky compilers happy
    return 0;
}

#else
#error "Unknown SYSTEM_TYPE"
#endif // SYSTEM_TYPE == WINDOWS
